# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-instance-attributes, too-many-lines

"""Compound configuration for Analog (Resistive Device) tiles."""

from copy import deepcopy
from dataclasses import dataclass, field
from typing import ClassVar, List, Type, Union, TYPE_CHECKING
from warnings import warn

from aihwkit.exceptions import ConfigError
from aihwkit.simulator.configs.helpers import (
    _PrintableMixin, parameters_to_bindings
)
from aihwkit.simulator.configs.utils import (
    IOParameters, UpdateParameters
)
from aihwkit.simulator.configs.enums import VectorUnitCellUpdatePolicy
from aihwkit.simulator.rpu_base import devices

if TYPE_CHECKING:
    from aihwkit.simulator.configs.devices import PulsedDevice


@dataclass
class UnitCell(_PrintableMixin):
    """Parameters that modify the behaviour of a unit cell."""

    bindings_class: ClassVar[Type] = devices.VectorResistiveDeviceParameter

    bindings_ignore: ClassVar[List] = ['diffusion', 'lifetime']

    unit_cell_devices: List = field(default_factory=list)
    """Devices that compose this unit cell."""

    construction_seed: int = 0
    """If not ``0``, set a unique seed for hidden parameters during
    construction.

    Applies to all ``unit_cell_devices``.
    """

    def as_bindings(self) -> devices.VectorResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        raise NotImplementedError

    def requires_diffusion(self) -> bool:
        """Return whether device has diffusion enabled."""
        return any(dev.requires_diffusion() for dev in self.unit_cell_devices)

    def requires_decay(self) -> bool:
        """Return whether device has decay enabled."""
        return any(dev.requires_decay() for dev in self.unit_cell_devices)


###############################################################################
# Specific devices based on ``unit cell``.
###############################################################################

@dataclass
class VectorUnitCell(UnitCell):
    """Abstract resistive device that combines multiple pulsed resistive
    devices in a single 'unit cell'.

    For instance, a vector device can consist of 2 resistive devices
    where the sum of the two resistive values are coded for each
    weight of a cross point.
    """

    bindings_class: ClassVar[Type] = devices.VectorResistiveDeviceParameter

    update_policy: VectorUnitCellUpdatePolicy = VectorUnitCellUpdatePolicy.ALL
    """The update policy of which if the devices will be receiving the update
    of a mini-batch."""

    first_update_idx: int = 0
    """Device that receives the first mini-batch.

    Useful only for ``VectorUnitCellUpdatePolicy.SINGLE_FIXED``.
    """

    gamma_vec: List[float] = field(default_factory=list, metadata={'hide_if': []})
    """Weighting of the unit cell devices to reduce to final weight.

    User-defined weightening can be given as a list if factors. If not
    given, each device index of the unit cell is weighted by equal
    amounts (:math:`1/n`).
    """

    def as_bindings(self) -> devices.VectorResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        vector_parameters = parameters_to_bindings(self)

        if not isinstance(self.unit_cell_devices, list):
            raise ConfigError('unit_cell_devices should be a list of devices')

        for param in self.unit_cell_devices:
            device_parameters = param.as_bindings()
            if not vector_parameters.append_parameter(device_parameters):
                raise ConfigError('Could not add unit cell device parameter')

        return vector_parameters


@dataclass
class ReferenceUnitCell(UnitCell):
    """Abstract device model takes two arbitrary device per cross-point and
    implements an device with reference pair.

    The update will only be on the 0-th device whereas the other will
    stay fixed. The resulting effective weight is the difference of
    the two.

    Note:
        Exactly 2 devices are used, if more are given the are
        discarded, if less, the same device will be used twice.

    Note:
        The reference device weights will all zero on default. To set
        the reference device with a particular value one can select the
        device update index::

            analog_tile.set_hidden_update_index(1)
            analog_tile.set_weights(W)
            analog_tile.set_hidden_update_index(0) # set back to 0 for the following updates
    """

    bindings_class: ClassVar[Type] = devices.VectorResistiveDeviceParameter

    update_policy: VectorUnitCellUpdatePolicy = VectorUnitCellUpdatePolicy.SINGLE_FIXED
    """The update policy of which if the devices will be receiving the
    update of a mini-batch.

    Caution:
        This parameter should be kept to SINGLE_FIXED for this device.
    """

    first_update_idx: int = 0
    """Device that receives the update."""

    gamma_vec: List[float] = field(default_factory=lambda: [1., -1.],
                                   metadata={'hide_if': [1., -1.]})
    """Weighting of the unit cell devices to reduce to final weight.

    Note:
        While user-defined weighting can be given it is suggested to keep it to
        the default ``[1, -1]`` to implement the reference device subtraction.
    """

    def as_bindings(self) -> devices.VectorResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        vector_parameters = parameters_to_bindings(self)

        if not isinstance(self.unit_cell_devices, list):
            raise ConfigError('unit_cell_devices should be a list of devices')

        if len(self.unit_cell_devices) > 2:
            self.unit_cell_devices = self.unit_cell_devices[:2]
        elif len(self.unit_cell_devices) == 1:
            self.unit_cell_devices = [self.unit_cell_devices[0],
                                      deepcopy(self.unit_cell_devices[0])]
        elif len(self.unit_cell_devices) != 2:
            raise ConfigError('ReferenceUnitCell expects two unit_cell_devices')

        for param in self.unit_cell_devices:
            device_parameters = param.as_bindings()
            if not vector_parameters.append_parameter(device_parameters):
                raise ConfigError('Could not add unit cell device parameter')

        return vector_parameters


@dataclass
class OneSidedUnitCell(UnitCell):
    """Abstract device model takes an arbitrary device per crosspoint and
    implements an explicit plus-minus device pair with one sided update.

    One device will receive all positive updated and the other all
    negative updates. Since the devices will quickly saturate, the
    device implements a refresh strategy.

    With fixed frequency per update call (``refresh_every``, in units
    of single vector updates) a refresh is performed. During the
    refresh, each column will be read using a forward pass (parameters
    are specified with ``refresh_forward``) to read out the positive and
    negative device weights.

    Whether a weight needs refreshing is determined by the following
    criterion: The larger weight (normalized by the tile-wise fixed
    w_max setting) is tested against the upper threshold. If larger
    than the upper threshold, and the normalized lower weight is
    larger than the lower threshold, then a reset and rewriting will
    be performed.

    Note that this abstract device needs single devices that are
    derived from :class:`~PulsedDevice`. The reset properties (bias
    and cycle-to-cycle noise) can be thus adjusted (see
    :class:`~PulsedDevice`).

    The rewriting of the computed difference is only done onto one of
    the two devices using the update properties defined in
    ``refresh_update``.

    Note:
        This device will take only the first ``unit_cell_device`` to
        generate two devices. Both positive and negative device will
        thus have the same (reversed) parameters, e.g. the specified
        ``w_min``, will become the w_max of the negative device.
    """

    bindings_class: ClassVar[Type] = devices.OneSidedResistiveDeviceParameter

    refresh_every: int = 0
    """How often a refresh is performed (in units of the number of vector
    updates).

    Note:
        If a refresh is done, full reads of both positive and negative
        devices are performed. Additionally, if single devices deemed
        to be refreshed, an (open-loop) re-write is done (once per
        column). Thus, refresh might have considerable runtime
        impacts.
    """

    units_in_mbatch: bool = True
    """If set, the ``refresh_every`` counter is given in ``m_batch``
    which is the re-use factor. Smaller numbers are not possible.

    Caution:
        For CUDA devices, refresh is always done in  ``m_batch`` (ie
        the number of re-use per layer for a mini-batch). Smaller
        numbers will have no effect.
    """

    refresh_upper_thres: float = 0.75
    """Upper threshold for determining the refresh, see above."""

    refresh_lower_thres: float = 0.25
    """Lower threshold for determining the refresh, see above."""

    refresh_forward: IOParameters = field(
        default_factory=IOParameters)
    """Input-output parameters that define the read during a refresh event.

    :class:`~aihwkit.simulator.config.utils.AnalogTileInputOutputParameters`
    that define the read (forward) of an refresh event. For instance
    the amount of noise or whether refresh is done using a ADC/DAC
    etc.
    """

    refresh_update: UpdateParameters = field(default_factory=UpdateParameters)
    """Update parameters that define the type of update used for each refresh
    event.

    Update parameters
    :class:`~aihwkit.simulator.config.utils.AnalogTileUpdateParameters`
    that define the type of update used for each refresh event.
    """

    copy_inverted: bool = False
    """Whether the use the "down" update behavior of the first device for
    the negative updates instead of the positive half of the second
    device."""

    def as_bindings(self) -> devices.OneSidedResistiveDeviceParameter:
        """Return a representation of this instance as a simulator
        bindings object."""
        if not isinstance(self.unit_cell_devices, list):
            raise ConfigError('unit_cell_devices should be a list of devices')

        onesided_parameters = parameters_to_bindings(self)
        device_parameter0 = self.unit_cell_devices[0].as_bindings()

        if len(self.unit_cell_devices) == 0 or len(self.unit_cell_devices) > 2:
            raise ConfigError('Need 1 or 2 unit_cell_devices')

        if len(self.unit_cell_devices) == 1:
            device_parameter1 = device_parameter0
        else:
            device_parameter1 = self.unit_cell_devices[1].as_bindings()

        # need to be exactly 2 and same parameters
        if not onesided_parameters.append_parameter(device_parameter0):
            raise ConfigError('Could not add unit cell device parameter')

        if not onesided_parameters.append_parameter(device_parameter1):
            raise ConfigError('Could not add unit cell device parameter ' +
                              '(both devices need to be of the same type)')

        return onesided_parameters


@dataclass
class DifferenceUnitCell(OneSidedUnitCell):
    """Deprecated alias to ``OneSidedUnitCell``."""

    def __post__init__(self) -> None:
        warn('The DifferenceUnitCell class is deprecated. Please use '
             'OneSidedUnitCell instead.',
             DeprecationWarning)


@dataclass
class TransferCompound(UnitCell):
    r"""Abstract device model that takes 2 or more devices and
    implements a transfer-based learning rule.

    It uses a (partly) hidden weight (where the SGD update is
    accumulated), which then is transferred partly and occasionally to
    the visible weight. This can implement an analog friendly variant
    of stochastic gradient descent (Tiki-taka), as described in
    `Gokmen & Haensch (2020)`_.

    The hidden weight is always the first in the list of
    ``unit_cell_devices`` given, and the transfer is done from left to
    right. The first of the ``unit_cell_devices`` can have different
    HW specifications from the rest, but the others need to be of
    identical specs. In detail, when specifying the list of devices
    only the first two will actually be used and the rest discarded
    and instead replaced by the second device specification. In this
    manner, the *fast* crossbar (receiving the SGD updates) and the
    *slow* crossbar (receiving the occasional partial transfers from
    the fast) can have different specs, but all additional slow
    crossbars (receiving transfers from the left neighboring crossbar
    in the list of ``unit_cell_devices``) need to be of the same spec.

    The rate of transfer (e.g. learning rate and how often and how
    many columns/rows per transfer) and the type (ie. with ADC or without,
    with noise etc.) can be adjusted.

    Each transfer event that is triggered by counting the update
    cycles (in units of either mini-batch or single mat-vecs),
    ``n_reads_per_transfer`` columns/rows are read from the left device
    using the forward pass with transfer vectors as input and
    transferred to the right (taking the order of the
    ``unit_cell_devices`` list) using the outer-product update with
    the read-out vectors and the transfer vectors. Currently, transfer
    vectors are fixed to be one-hot vectors. The columns/rows to take are
    in sequential order and warped around at the edge of the
    crossbar. The learning rate and forward and update specs of the
    transfer can be user-defined.

    The weight that is seen in the forward and backward pass is
    governed by the :math:`\gamma` weightening setting.

    Note:
        Here the devices could be either transferred in analog
        (essentially within the unit cell) or on separate arrays (using
        the usual (non-ideal) forward pass and update steps. This can be
        set with ``transfer_forward`` and ``transfer_update``.

    .. _Gokmen & Haensch (2020): https://www.frontiersin.org/articles/10.3389/fnins.2020.00103/full
    """

    bindings_class: ClassVar[Type] = devices.TransferResistiveDeviceParameter

    gamma: float = 0.0
    r"""Weighting factor to compute the effective SGD weight from the hidden
    matrices.

    The default scheme is:

    .. math:: g^{n-1} W_0 + g^{n-2} W_1 + \ldots + g^0  W_{n-1}
    """

    gamma_vec: List[float] = field(default_factory=list,
                                   metadata={'hide_if': []})
    """User-defined weightening.

    User-defined weightening can be given as a list if weights in which case
    the default weightening scheme with ``gamma`` is not used.
    """

    transfer_every: float = 1.0
    """Transfers every :math:`n` mat-vec operations or :math:`n` batches.

    Transfers every :math:`n` mat-vec operations (rounded to multiples/ratios
    of ``m_batch`` for CUDA). If ``units_in_mbatch`` is set, then the units are
    in ``m_batch`` instead of mat-vecs, which is equal to the overall the
    weight re-use during a while mini-batch.

    Note:
        If ``transfer_every`` is 0.0 *no transfer* will be made.

    If not given explicitely with ``transfer_every_vec``, then the higher
    transfer cycles are geometrically scaled, the first is set to
    transfer_every. Each next transfer cycle is multiplied by ``x_size
    / n_reads_per_transfer``.
    """

    no_self_transfer: bool = True
    """Whether to set the transfer rate of the last device (which is applied to
    itself) to zero."""

    transfer_every_vec: List[float] = field(default_factory=list,
                                            metadata={'hide_if': []})
    """Transfer cycles lengths.

    A list of :math:`n` entries, to explicitly set the transfer cycles lengths.
    In this case, the above defaults are ignored.
    """

    units_in_mbatch: bool = True
    """Units for ``transfer_every``.

    If set, then the cycle length units of ``transfer_every`` are in
    ``m_batch`` instead of mat-vecs, which is equal to the overall of the
    weight re-use during a while mini-batch.
    """

    n_reads_per_transfer: int = 1
    """Number of consecutive reads to use during transfer events.

    How many consecutive columns or rows to read (from one tile) and write (to the next
    tile) every transfer event. For read, the input is a 1-hot vector. Once the
    final columns or row is reached, reading starts again from the first.
    """

    transfer_columns: bool = True
    """Whether to read and transfer columns or rows.

    If set, read is done with an additional forward pass
    determined by the ``transfer_forward`` settings. If not set, rows
    are transferred instead, that is, the read is done internally
    with a backward pass instead. However, the parameters defining the
    backward are still given by setting the ``transfer_forward`` field for
    convenience.
    """

    with_reset_prob: float = 0.0
    """Whether to apply reset of the columns that were transferred with a given
    probability.

    Note:
        Reset is only available in case of column reads
        (``transfer_columns==True``).
    """

    random_selection: bool = False
    """Whether to select a random starting column or row.

    Whether to select a random starting column or row for each
    transfer event and not take the next column or row that was
    previously not transferred as a starting column or row (the
    default).
    """

    fast_lr: float = 0.0
    """Whether to set the `fast` tile's learning rate.

    If set, then the SGD gradient update onto the first (fast) tile is
    set to this learning rate and is kept constant even when the SGD
    learning rate is scheduled. The SGD learning rate is then only
    used to scale the transfer LR (see ``scale_transfer_lr``).
    """

    transfer_lr: float = 1.0
    """Learning rate (LR) for the update step of the transfer event.

    Per default all learning rates are identical. If ``scale_transfer_lr`` is
    set, the transfer LR is scaled by current learning rate of the SGD.

    Note:
        LR is always a positive number, sign will be correctly
        applied internally.
    """

    transfer_lr_vec: List[float] = field(default_factory=list,
                                         metadata={'hide_if': []})
    """Transfer LR for each individual transfer in the device chain can be
    given."""

    scale_transfer_lr: bool = True
    """Whether to give the transfer_lr in relative units.

    ie. whether to scale the transfer LR with the current LR of the SGD.
    """

    transfer_forward: IOParameters = field(
        default_factory=IOParameters)
    """Input-output parameters that define the read of a transfer event.

    :class:`~aihwkit.simulator.config.utils.AnalogTileInputOutputParameters` that define the read
    (forward or backward) of an transfer event. For instance the amount of noise
    or whether transfer is done using a ADC/DAC etc.
    """

    transfer_update: UpdateParameters = field(
        default_factory=UpdateParameters)
    """Update parameters that define the type of update used for each transfer
    event.

    Update parameters :class:`~aihwkit.simulator.config.utils.AnalogTileUpdateParameters` that
    define the type of update used for each transfer event.
    """

    def as_bindings(self) -> devices.TransferResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        if not isinstance(self.unit_cell_devices, list):
            raise ConfigError('unit_cell_devices should be a list of devices')

        n_devices = len(self.unit_cell_devices)

        transfer_parameters = parameters_to_bindings(self)

        param_fast = self.unit_cell_devices[0].as_bindings()
        param_slow = self.unit_cell_devices[1].as_bindings()

        if not transfer_parameters.append_parameter(param_fast):
            raise ConfigError('Could not add unit cell device parameter')

        for _ in range(n_devices - 1):
            if not transfer_parameters.append_parameter(param_slow):
                raise ConfigError('Could not add unit cell device parameter')

        return transfer_parameters


@dataclass
class BufferedTransferCompound(TransferCompound):
    r"""Abstract device model that takes 2 or more devices and
    implements a buffered transfer-based learning rule.

    Different to :class:`TransferCompound`, however,  readout is done
    first onto a digital buffer (in floating point precision), from
    which then the second analog matrix is updated. This second step is
    very similar to the analog update in :class:`MixedPrecisionCompound`.

    Note, however, that in contrast to :class:`MixedPrecisionCompound`
    the rank-update is still done in analog with parallel update using
    pulse trains.

    The buffer is assumed to be in floating point precision and
    only one row/column at a time needs to be processed in one update cycle,
    thus greatly reducing on-chip memory requirements.

    For details, see `Gokmen (2021)`_.

    .. _Gokmen (2021): https://www.frontiersin.org/articles/10.3389/frai.2021.699148/full
    """

    bindings_class: ClassVar[Type] = devices.BufferedTransferResistiveDeviceParameter

    thres_scale: float = 1.0
    """Threshold scale for buffer to determine whether to transfer to next
    device. Will be multiplied by the device granularity to get the
    threshold.
    """

    step: float = 1.0
    """Value to fill the ``d`` vector for the update if buffered value is
    above threshold.
    """

    momentum: float = 0.1
    """Momentum of the buffer.

    After transfer, this momentum fraction stays on the buffer instead
    of subtracting all of what was transferred.
    """

    forget_buffer: bool = True
    """Whether to forget the value of the buffer after transfer.

    If enabled, the buffer is reset to the momentum times the
    transferred value. Thus, if the number of pulses is limited to
    e.g. 1 (``desired_BL`` in the ``transfer_update``) the transfer
    might be clipped and the potentially larger buffer values are
    forgotten. If disabled, then the buffer values are faithfully
    subtracted by the amount transferred (times one minus momentum).
    """

    transfer_update: UpdateParameters = field(
        default_factory=lambda: UpdateParameters(desired_bl=1, update_bl_management=False,
                                                 update_management=False))
    """Update parameters that define the type of update used for each transfer
    event.

    Update parameters :class:`~aihwkit.simulator.config.utils.AnalogTileUpdateParameters` that
    define the type of update used for each transfer event.
    """


###############################################################################
# Specific compound-devices with digital rank update
###############################################################################

@dataclass
class DigitalRankUpdateCell(_PrintableMixin):
    """Parameters that modify the behavior of the digital rank update cell.

    This is the base class for devices that compute the rank update in
    digital and then (occasionally) transfer the information to the
    (analog) crossbar array that is used during forward and backward.
    """

    bindings_class: ClassVar[Type] = devices.AbstractResistiveDeviceParameter

    bindings_ignore: ClassVar[List] = ['diffusion', 'lifetime']

    device: Union['PulsedDevice',
                  OneSidedUnitCell,
                  VectorUnitCell,
                  ReferenceUnitCell] = field(
                      default_factory=VectorUnitCell)
    """(Analog) device that are used for forward and backward."""

    construction_seed: int = 0
    """If not ``0``, set a unique seed for hidden parameters during
    construction.

    Applies to ``device``.
    """

    def as_bindings(self) -> devices.AbstractResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        raise NotImplementedError

    def requires_diffusion(self) -> bool:
        """Return whether device has diffusion enabled."""
        return self.device.requires_diffusion()

    def requires_decay(self) -> bool:
        """Return whether device has decay enabled."""
        return self.device.requires_decay()


@dataclass
class MixedPrecisionCompound(DigitalRankUpdateCell):
    r"""Abstract device model that takes 1 (analog) device and
    implements a transfer-based learning rule, where the outer product
    is computed in digital.

    Here, the outer product of the activations and error is done on a
    full-precision floating-point :math:`\chi` matrix. Then, with a
    threshold given by the ``granularity``, pulses will be applied to
    transfer the information row-by-row to the analog matrix.

    For details, see `Nandakumar et al. Front. in Neurosci. (2020)`_.

    Note:
        This version of update is different from a parallel update in
        analog other devices are implementing with stochastic pulsing,
        as here :math:`{\cal O}(n^2)` digital computations are needed
        to compute the outer product (rank update). This need for
        digital compute in potentially high precision might result in
        inferior run time and power estimates in real-world
        applications, although sparse integer products can potentially
        be employed to speed up to improve run time estimates. For
        details, see discussion in `Nandakumar et al. Front. in
        Neurosci. (2020)`_.

    .. _`Nandakumar et al. Front. in Neurosci. (2020)`: https://doi.org/10.3389/fnins.2020.00406
    """

    bindings_class: ClassVar[Type] = devices.MixedPrecResistiveDeviceParameter

    transfer_every: int = 1
    """Transfers every :math:`n` mat-vec operations.
    Transfers every :math:`n` mat-vec operations (rounded to multiples/ratios
    of ``m_batch``).

    Standard setting is 1.0 for mixed precision, but it could potentially be
    reduced to get better run time estimates.
    """

    n_rows_per_transfer: int = -1
    r"""How many consecutive rows to write to the tile from the :math:`\chi`
    matrix.

    ``-1`` means full matrix read each transfer event.
    """

    random_row: bool = False
    """Whether to select a random starting row.

    Whether to select a random starting row for each transfer event and not
    take the next row that was previously not transferred as a starting row
    (the default).
    """

    granularity: float = 0.0
    r"""Granularity of the device.

    Granularity :math:`\varepsilon` of the device that is used to
    calculate the number of pulses transferred from :math:`\chi` to
    analog.

    If 0, it will take granularity from the analog device used.
    """

    transfer_lr: float = 1.0
    r"""Scale of the transfer to analog .

    The update onto the analog tile will be proportional to
    :math:`\langle\chi/\varepsilon\rangle\varepsilon\lambda_\text{tr}`,
    where :math:`\lambda_\text{tr}` is given by ``transfer_lr`` and
    :math:`\varepsilon` is the granularity.
    """

    n_x_bins: int = 0
    """The number of bins to discretize (symmetrically around zero) the
    activation before computing the outer product.

    Dynamic quantization is used by computing the absolute max value of each
    input. Quantization can be turned off by setting this to 0.
    """

    n_d_bins: int = 0
    """The number of bins to discretize (symmetrically around zero) the
    error before computing the outer product.

    Dynamic quantization is used by computing the absolute max value of each
    error vector. Quantization can be turned off by setting this to 0.
    """

    stoc_round_x: bool = True
    """Whether to use stochastic rounding in case of quantization of the input x.
    """

    stoc_round_d: bool = True
    """Whether to use stochastic rounding in case of quantization of the error d.
    """

    def as_bindings(self) -> devices.MixedPrecResistiveDeviceParameter:
        """Return a representation of this instance as a simulator bindings object."""
        mixed_prec_parameter = parameters_to_bindings(self)
        param_device = self.device.as_bindings()

        if not mixed_prec_parameter.set_device_parameter(param_device):
            raise ConfigError('Could not add device parameter')

        return mixed_prec_parameter



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""aihwkit example 1: simple network with one layer.
Simple network that consist of one analog layer. The network aims to learn
to sum all the elements from one array.
"""
# pylint: disable=invalid-name, too-many-locals

from tempfile import TemporaryFile
from numpy import array

# Imports from PyTorch.
from torch import randn, load, save, manual_seed
from torch.nn.functional import mse_loss

# Imports from aihwkit.
from aihwkit.nn import (
    AnalogLinearMapped, AnalogConv1dMapped,
    AnalogConv2dMapped,  AnalogConv3dMapped
)
from aihwkit.optim import AnalogSGD

from .helpers.decorators import parametrize_over_layers
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint, Inference, Ideal
from .helpers.layers import (
    Linear, LinearCuda, Conv1d, Conv1dCuda, Conv2d,
    Conv2dCuda, Conv3dCuda
)

DECIMAL = 4


@parametrize_over_layers(
    layers=[Linear, LinearCuda, Conv1d, Conv2d, Conv1dCuda, Conv2dCuda, Conv3dCuda],
    tiles=[FloatingPoint, Inference, Ideal],
    biases=['digital', None])
class MappedLayerLinearTest(ParametrizedTestCase):
    """Tests for the AnalogMappedLayer functionality """

    def get_mapped_class(self, model):
        """Returns the mapped class"""
        name = model.__class__.__name__
        if name == 'AnalogLinear':
            return AnalogLinearMapped
        if name == 'AnalogConv1d':
            return AnalogConv1dMapped
        if name == 'AnalogConv2d':
            return AnalogConv2dMapped
        if name == 'AnalogConv3d':
            return AnalogConv3dMapped

        raise RuntimeError("Cannot find mapped module.")

    def get_mapped_model(self, model, rpu_config):
        """Returns the mapped model"""
        mapped_model = self.get_mapped_class(model).from_digital(
            model, rpu_config=rpu_config)
        mapped_model.reset_parameters()  # should set it explicitly below
        return mapped_model

    def get_image_size(self, model):
        """Returns the image size"""
        if not hasattr(model, 'kernel_size'):
            return []
        return (array(model.kernel_size) * 2).tolist()

    def train_model(self, model, in_vectors, out_vectors):
        """Trains a model """

        opt = AnalogSGD(model.parameters(), lr=0.1)

        for _ in range(10):
            opt.zero_grad()

            # Add the training Tensor to the model (input).
            pred_value = model(in_vectors)
            # Add the expected output Tensor.
            loss_value = mse_loss(pred_value, out_vectors)
            # Run training (backward propagation).
            loss_value.backward()
            opt.step()

        return loss_value

    def test_construction(self):
        """ Test construction of a mapped layer"""
        manual_seed(123)
        in_features = 14
        out_features = 5

        rpu_config = self.get_rpu_config()
        rpu_config.mapping.max_input_size = 10
        rpu_config.mapping.max_output_size = 4

        model = self.get_layer(in_features, out_features, rpu_config=rpu_config)
        weight, bias = model.get_weights()

        weight = randn(*weight.shape)
        if self.bias:
            bias = randn(*bias.shape)
        model.set_weights(weight, bias)

        mapped_model = self.get_mapped_model(model, rpu_config)
        mapped_model.set_weights(weight, bias)

        if self.use_cuda:
            mapped_model = mapped_model.cuda()

        mapped_weights, mapped_bias = mapped_model.get_weights()

        self.assertTensorAlmostEqual(weight, mapped_weights, decimal=DECIMAL)
        if self.bias:
            self.assertTensorAlmostEqual(bias, mapped_bias, decimal=DECIMAL)

    def test_training(self):
        """ Test of training of a mapped linear layer"""
        manual_seed(123)

        in_features = 12
        out_features = 13
        batch_size = 10

        rpu_config = self.get_rpu_config()
        rpu_config.mapping.max_input_size = 10
        rpu_config.mapping.max_output_size = 6

        model = self.get_layer(in_features, out_features, rpu_config=rpu_config)
        weight, bias = model.get_weights()

        weight = randn(*weight.shape)
        if self.bias:
            bias = randn(*bias.shape)
        model.set_weights(weight, bias)

        mapped_model = self.get_mapped_model(model, rpu_config)
        mapped_model.set_weights(weight, bias)

        in_vectors = randn(*([batch_size, in_features] + self.get_image_size(model)))

        if self.use_cuda:
            in_vectors = in_vectors.cuda()
            mapped_model = mapped_model.cuda()

        out_vectors = randn(*model(in_vectors).shape)
        if self.use_cuda:
            out_vectors = out_vectors.cuda()

        # compare predictions for analog linear and analog spit linear layers
        self.assertTensorAlmostEqual(model(in_vectors), mapped_model(in_vectors), decimal=DECIMAL)

        # Define an analog-aware optimizer, preparing it for using the layers.
        loss = self.train_model(model, in_vectors, out_vectors)
        mapped_loss = self.train_model(mapped_model, in_vectors, out_vectors)

        self.assertTensorAlmostEqual(loss, mapped_loss, decimal=DECIMAL)

        # Make sure that the train model produces the same forward pass
        self.assertTensorAlmostEqual(model(in_vectors), mapped_model(in_vectors), decimal=DECIMAL)

    def test_training_after_save(self):
        """ Test training after it was saved """
        manual_seed(123)

        in_features = 11
        out_features = 11
        batch_size = 10

        rpu_config = self.get_rpu_config()
        rpu_config.mapping.max_input_size = 10
        rpu_config.mapping.max_output_size = 4

        model = self.get_layer(in_features, out_features, rpu_config=rpu_config)
        weight, bias = model.get_weights()

        weight = randn(*weight.shape)
        if self.bias:
            bias = randn(*bias.shape)
        model.set_weights(weight, bias)

        mapped_model = self.get_mapped_model(model, rpu_config)
        mapped_model.set_weights(weight, bias)

        in_vectors = randn(*([batch_size, in_features] + self.get_image_size(model)))
        if self.use_cuda:
            in_vectors = in_vectors.cuda()
            mapped_model = mapped_model.cuda()

        out_vectors = randn(*model(in_vectors).shape)
        if self.use_cuda:
            out_vectors = out_vectors.cuda()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(mapped_model.state_dict(), file)
            # Create a new model and load its state dict.
            file.seek(0)
            new_model = self.get_mapped_model(model, rpu_config)
            if self.use_cuda:
                new_model = new_model.cuda()
            new_model.load_state_dict(load(file))

        new_weight, new_bias = new_model.get_weights()

        self.assertTensorAlmostEqual(new_weight, weight, decimal=DECIMAL)
        if self.bias:
            self.assertTensorAlmostEqual(new_bias, bias, decimal=DECIMAL)
            self.assertTensorAlmostEqual(model.bias, bias, decimal=DECIMAL)
            self.assertTensorAlmostEqual(mapped_model.bias, bias, decimal=DECIMAL)

        for new_tile, tile in zip(list(new_model.analog_tiles()),
                                  list(mapped_model.analog_tiles())):
            new_tile_weight, _ = new_tile.get_weights()
            tile_weight, _ = tile.get_weights()
            self.assertTensorAlmostEqual(tile_weight, new_tile_weight, decimal=DECIMAL)

        # compare predictions for analog linear and analog spit linear layers
        self.assertTensorAlmostEqual(model(in_vectors), new_model(in_vectors), decimal=DECIMAL)

        # Define an analog-aware optimizer, preparing it for using the layers.
        loss = self.train_model(model, in_vectors, out_vectors)
        new_loss = self.train_model(new_model, in_vectors, out_vectors)

        # compare predictions for analog linear and analog spit linear layers
        self.assertTensorAlmostEqual(model(in_vectors), new_model(in_vectors), decimal=DECIMAL)

        self.assertTensorAlmostEqual(loss, new_loss, decimal=DECIMAL)

        # Make sure that the train model produces the same forward pass
        self.assertTensorAlmostEqual(model(in_vectors), new_model(in_vectors), decimal=DECIMAL)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Creates InferenceRPUConfig to add to nn model"""
from typing import Dict, Any
from collections import OrderedDict

from aihwkit.simulator.configs.configs import InferenceRPUConfig
from aihwkit.simulator.presets.web import (
    WebComposerInferenceRPUConfig,
    OldWebComposerInferenceRPUConfig
)
from aihwkit.inference.noise.pcm import PCMLikeNoiseModel
from aihwkit.inference.noise.custom import StateIndependentNoiseModel
from aihwkit.inference.compensation.drift import GlobalDriftCompensation
from aihwkit.cloud.converter.v1.analog_info import AnalogInfo
from aihwkit.cloud.converter.v1.noise_model_info import NoiseModelInfo

RPU_CLASSES = {
    'InferenceRPUConfig': InferenceRPUConfig,
    'WebComposerInferenceRPUConfig': WebComposerInferenceRPUConfig,
    'OldWebComposerInferenceRPUConfig': OldWebComposerInferenceRPUConfig
}


# pylint: disable=too-few-public-methods
class NoiseModelDeviceIDException(Exception):
    """Exception raised if noise model device id is not correct"""


class RPUconfigInfo:
    """Data only class for RPUConfig fields"""

    def __init__(self, nm_info: NoiseModelInfo,
                 a_info: AnalogInfo,
                 layers: Any = None):
        """
        The only constructor for this class
        """
        self._noise_model_info = nm_info
        self._analog_info = a_info
        self._layers = layers
        self._device_id = ''

    @staticmethod
    def _get_common_rpucfg_name(layers: Any) -> Any:
        """Set common rpu config name by search all analog layers"""
        # Use default RPU config for Composer
        if layers is None:
            return 'WebComposerInferenceRPUConfig'
        # Need to loop through protobuf layers and figure out
        #   common rpu_config value.
        names: Dict[str, int] = {}
        # pylint: disable=too-many-nested-blocks
        for layer_proto in layers:  # type: ignore[attr-defined]
            if layer_proto.WhichOneof('item') == 'layer':
                layer = layer_proto.layer
                if layer.id.startswith('Analog'):
                    # Loop though all AttributeProto objecs in layer.arguments
                    for argument in layer.arguments:
                        if argument.name == 'rpu_config':
                            # stored as UTF8 byte string in attribute s
                            arg_value = getattr(argument, 's')
                            # update count of this rpu_config in all analog layers
                            if arg_value in names:
                                names[arg_value] += 1
                            else:
                                names[arg_value] = 1
        # pylint: enable=too-many-nested-blocks
        # should have exactly on in dictionary 'names'
        if len(names) > 1:
            print(f'>>> ERROR: more than one rpu_config: {names}')
            return None
        if len(names) == 1:
            # keys() returns dict_keys object, need a list
            return list(names.keys())[0].decode('UTF-8')  # type: ignore[attr-defined]
        print('>>> INFO: experiment has not analog layers')
        return ''

    def _print_rpu_config(
            self,
            rpu_config: InferenceRPUConfig,
            func_id: str,
            dev_id: str) -> None:
        """Pretty-print rpu_config"""

        # Create ordered dictionary with field name as key and value as field value
        #   Include ADC and DAC in input/output resolution
        print_order = OrderedDict(
            {
                ('forward.out_noise', rpu_config.forward.out_noise),
                ('forward.out_res/adc', '{}/{}'.format(
                    rpu_config.forward.out_res, self._analog_info.adc)),
                ('forward.inp_res/dac', '{}/{}'.format(
                    rpu_config.forward.inp_res, self._analog_info.dac)),
                ('noise_model', rpu_config.noise_model),
                ('noise_model.read_noise_scale',
                    rpu_config.noise_model.read_noise_scale),  # type: ignore[attr-defined]
                ('noise_model.prog_noise_scale',
                    rpu_config.noise_model.prog_noise_scale),  # type: ignore[attr-defined]
                ('drift_compensation',
                    rpu_config.drift_compensation),  # type: ignore[attr-defined]
                ('noise_model.drift_scale',
                    rpu_config.noise_model.drift_scale),  # type: ignore[attr-defined]
            }

        )

        # add extra fields for GENERIC device
        if dev_id == NoiseModelInfo.GENERIC:
            print_order['noise_model.drift_nu_mean'] = (
                rpu_config.noise_model.drift_nu_mean)  # type: ignore[attr-defined]
            print_order['noise_model.drift_nu_std'] = (
                rpu_config.noise_model.drift_nu_std)  # type: ignore[attr-defined]

        output = '  rpu_config: function_id={} device_id={}\n'.format(func_id, dev_id)
        for key in print_order:
            output += '    {:30}: {}\n'.format(key, print_order[key])
            if key == 'noise_model' and dev_id == NoiseModelInfo.PCM:
                # need to print out prog_coeff in PCMLikeNoiseModel object
                for i in range(3):
                    field_name = 'noise_model.prog_coeff[{}]'.format(i)
                    output += '    {:30}: {}\n'.format(
                        field_name,
                        rpu_config.noise_model.prog_coeff[i]  # type: ignore[attr-defined]
                    )

        print('Here is the variable content of the rpu_config:\n{}'.format(output))

    def create_inference_rpu_config(self, func_id: str,
                                    verbose: bool = False) -> InferenceRPUConfig:
        """Creates a InferenceRPUConfig class using noise and analog info"""
        # Need to find name of 'common-rpu-conf-class-name' in protobuf
        #   This should be the consistent across all layers.
        #   The Composer Validator should have already caught this but
        #   it is checked here for testcases and other unknown environments
        rpu_class_name = self._get_common_rpucfg_name(self._layers)
        print(f'>>> INFO: rpu_class_name={rpu_class_name}')
        if rpu_class_name is None or len(rpu_class_name) == 0:
            raise Exception('class name error. see previous messages')
        rpu_config_class = None
        if rpu_class_name in RPU_CLASSES:
            rpu_config_class = RPU_CLASSES[rpu_class_name]
        else:
            raise Exception(f"rpu class name '{rpu_class_name}' not one of '{RPU_CLASSES.keys()}'")

        # Dynamically create the right InferenceRPUConfig class
        rpu_config = rpu_config_class()
        # Assign values from AnalogProto
        rpu_config.forward.out_noise = self._analog_info.output_noise_strength

        # changed input/output res to use a formula. print out adc and dac
        rpu_config.forward.out_res = 1.0 / (2**self._analog_info.adc - 2)
        rpu_config.forward.inp_res = 1.0 / (2**self._analog_info.dac - 2)

        # Assign values from NoiseModelProto (CM Noise model)
        self._device_id = self._noise_model_info.device_id
        if self._device_id == NoiseModelInfo.PCM:
            rpu_config.noise_model = (
                PCMLikeNoiseModel(
                                g_max=25.0,
                                prog_coeff=[
                                    self._noise_model_info.poly_constant_coef,
                                    self._noise_model_info.poly_first_order_coef,
                                    self._noise_model_info.poly_second_order_coef,
                                ]
                )
            )
        elif self._device_id == NoiseModelInfo.GENERIC:
            rpu_config.noise_model = StateIndependentNoiseModel(
                                g_max=25.0,
                                prog_coeff=[
                                    self._noise_model_info.poly_constant_coef,
                                    self._noise_model_info.poly_first_order_coef,
                                    self._noise_model_info.poly_second_order_coef,
                                ]
            )

            # These are unique to generic
            rpu_config.noise_model.drift_nu_mean = (
                self._noise_model_info.drift_mean  # type: ignore[attr-defined]
            )
            rpu_config.noise_model.drift_nu_std = (
                self._noise_model_info.drift_std  # type: ignore[attr-defined]
            )
        else:
            raise NoiseModelDeviceIDException(
                'invalid noise model device id {}'.format(self._device_id))

        # common fields
        rpu_config.noise_model.prog_noise_scale = (  # type: ignore[attr-defined]
            self._noise_model_info.programming_noise_scale
        )
        rpu_config.noise_model.read_noise_scale = (  # type: ignore[attr-defined]
                self._noise_model_info.read_noise_scale
        )

        # Drift compensation in protobuf is boolean (bool)
        rpu_config.drift_compensation = None  # type: ignore[assignment]
        if self._noise_model_info.drift_compensation:
            rpu_config.drift_compensation = GlobalDriftCompensation()

        rpu_config.noise_model.drift_scale = (  # type: ignore[attr-defined]
                self._noise_model_info.drift_scale
        )

        if verbose:
            self._print_rpu_config(rpu_config, func_id, self._noise_model_info.device_id)
        return rpu_config



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (base)."""
# pylint: disable=too-many-lines, wrong-import-position

from collections import OrderedDict
from typing import (
    Dict, Generic, List, Optional,
    Tuple, TypeVar, Union, TYPE_CHECKING, Any
)
from copy import deepcopy
from numpy.typing import ArrayLike
from numpy import array

from torch import (
    Tensor, stack, zeros, as_tensor, cat,
    unsqueeze, squeeze, ones,
    float32, from_numpy, full, clamp,
    zeros_like, eye, randn,
)

from torch import device as torch_device
from torch import max as torch_max
from torch.nn import Parameter
from torch.autograd import no_grad
from torch.linalg import lstsq

from aihwkit.simulator.rpu_base import tiles
from aihwkit.exceptions import TileError, ConfigError
from aihwkit.optim.context import AnalogContext

RPUConfigGeneric = TypeVar('RPUConfigGeneric')

if TYPE_CHECKING:
    from aihwkit.simulator.configs.utils import MappingParameter, InputRangeParameter


class AnalogTileStateNames:  # pylint: disable=too-few-public-methods
    """ Class defining analog tile state name constants.

    Caution:
       Do *not* edit. Some names are attribute names of the tile.
    """

    WEIGHTS = 'analog_tile_weights'
    HIDDEN_PARAMETERS = 'analog_tile_hidden_parameters'
    HIDDEN_PARAMETER_NAMES = 'analog_tile_hidden_parameter_names'
    CLASS = 'analog_tile_class'
    LR = 'analog_lr'
    SHARED_WEIGHTS = 'shared_weights'
    CONTEXT = 'analog_ctx'
    OUT_SCALING = 'out_scaling_alpha'
    MAPPING_SCALES = 'mapping_scales'
    RPU_CONFIG = 'rpu_config'


class BaseTile(Generic[RPUConfigGeneric]):
    """Base class for tiles.

    Args:
        out_size: output size
        in_size: input size
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile.
        in_trans: Whether to assume an transposed input (batch first)
        out_trans: Whether to assume an transposed output (batch first)
    """
    # pylint: disable=too-many-instance-attributes,too-many-public-methods

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: RPUConfigGeneric,
            bias: bool = True,
            in_trans: bool = False,
            out_trans: bool = False
    ):
        self.out_size = out_size
        self.in_size = in_size
        self.rpu_config = deepcopy(rpu_config)
        self.bias = bias
        self.in_trans = in_trans
        self.out_trans = out_trans
        self.shared_weights = None  # type: Parameter
        self.out_scaling_alpha = None  # type: Parameter
        self.mapping_scales = None  # type: Tensor
        self.input_range = None  # type: Parameter

        # Whether CUDA-calls should be blocking
        self.non_blocking = False

        # Only used for indexed.
        self.image_sizes = []  # type: List[int]

        x_size = in_size + 1 if self.bias else in_size
        d_size = out_size

        self.tile = self._create_simulator_tile(x_size, d_size, rpu_config)
        self.tile.set_learning_rate(0.01)
        self.tile.set_weights_uniform_random(-0.01, 0.01)

        self.device = torch_device('cpu')
        self.is_cuda = False

        # create analog context
        self.analog_ctx = AnalogContext(self)

        # Helpers.
        self.reference_combined_weights = None  # type: Optional[Tensor]

        # init input / output processing
        self.init_learned_out_scales()
        self.init_mapping_scales()
        self.init_input_processing()

    @no_grad()
    def get_analog_ctx(self) -> AnalogContext:
        """Return the analog context of the tile to be used in ``AnalogFunction``."""
        return self.analog_ctx

    @no_grad()
    def ensure_shared_weights(self, shared_weights: Optional[Tensor] = None) -> None:
        """Ensure that the shared_weights is set properly.

        Caution:
           This is only called from analog function.

        No-op if shared weights is not used.
        """
        if shared_weights is not None:
            self.shared_weights.data = shared_weights.data  # type: ignore

        if self.shared_weights is not None:
            self.tile.set_shared_weights(self.shared_weights.data)

    @no_grad()
    def set_delta_weights(self, delta_weights: Optional[Tensor] = None) -> None:
        """Set the weight grad tensor and set the update to.

        No-op if shared weights is not used.
        """
        if self.shared_weights is not None and delta_weights is not None:
            self.tile.set_delta_weights(delta_weights)

    @no_grad()
    def reset_delta_weights(self) -> None:
        """Reset the weight grad tensor to default update behavior (i.e. adding the
        update directly to the weight).

        No-op if shared weights is not used.
        """
        if self.shared_weights is not None:
            self.tile.reset_delta_weights()

    @no_grad()
    def get_brief_info(self) -> str:
        """Return short info about the underlying C++ tile."""
        return self.tile.get_brief_info().rstrip()

    def __getstate__(self) -> Dict:
        """Get the state for pickling.

        This method removes the ``tile`` member, as the binding Tiles are not
        serializable.
        """
        # Caution: all attributes of the tile will be saved.
        current_dict = self.__dict__.copy()

        SN = AnalogTileStateNames
        current_dict[SN.WEIGHTS] = self.tile.get_weights()
        current_dict[SN.HIDDEN_PARAMETERS] \
            = self.tile.get_hidden_parameters().data
        current_dict[SN.HIDDEN_PARAMETER_NAMES] \
            = self.tile.get_hidden_parameter_names()
        current_dict[SN.CLASS] = self.__class__.__name__
        current_dict[SN.LR] = self.tile.get_learning_rate()
        current_dict.pop('tile', None)

        # don't save device. Will be determined by loading object
        current_dict.pop('stream', None)
        current_dict.pop('is_cuda', None)
        current_dict.pop('device', None)

        # this is should not be saved.
        current_dict.pop('image_sizes', None)

        return current_dict

    def __setstate__(self, state: Dict) -> None:
        """Set the state after unpickling.

        This method recreates the ``tile`` member, creating a new one from
        scratch, as the binding Tiles are not serializable.

        Caution:
            RPU configs are overwritten by loading the state.

        Raises:
            TileError: if tile class does not match or hidden parameters do not match
        """
        # pylint: disable=too-many-locals, too-many-statements, too-many-branches

        # Note: self here is NOT initialized! So we need to recreate
        # attributes that were not saved in getstate
        SN = AnalogTileStateNames
        current_dict = state.copy()
        current_dict.pop('image_sizes', None)  # should not be saved
        weights = current_dict.pop(SN.WEIGHTS)

        hidden_parameters = current_dict.pop(SN.HIDDEN_PARAMETERS)
        hidden_parameters_names = current_dict.pop(SN.HIDDEN_PARAMETER_NAMES, [])
        alpha_scale = current_dict.pop('analog_alpha_scale', None)  # legacy
        tile_class = current_dict.pop(SN.CLASS, self.__class__.__name__)
        analog_lr = current_dict.pop(SN.LR, 0.01)
        analog_ctx = current_dict.pop(SN.CONTEXT)
        shared_weights = current_dict.pop(SN.SHARED_WEIGHTS)
        shared_weights_if = shared_weights is not None

        mapping_scales = current_dict.pop(SN.MAPPING_SCALES, None)
        learned_out_scales = current_dict.pop(SN.OUT_SCALING, None)

        current_dict.pop('noise_model', None)  # legacy
        current_dict.pop('drift_compensation', None)  # legacy

        # legacy
        if 'non_blocking' not in current_dict:
            current_dict['non_blocking'] = False

        self.__dict__.update(current_dict)

        self.device = torch_device('cpu')
        self.is_cuda = False
        # get the current map location from analog_ctx (which is restored)
        to_device = analog_ctx.device

        # recreate attributes not saved
        # always first create on CPU
        x_size = self.in_size + 1 if self.bias else self.in_size
        d_size = self.out_size

        # Recreate the tile.
        # Check for tile mismatch
        if tile_class != self.__class__.__name__:
            raise TileError(
                'Mismatch of tile class: {} versus {}. Can only load analog '
                'state from the same tile class.'.format(self.__class__.__name__, tile_class))

        self.tile = self._create_simulator_tile(x_size, d_size, self.rpu_config)
        names = self.tile.get_hidden_parameter_names()
        if len(hidden_parameters_names) > 0 and names != hidden_parameters_names:
            # Check whether names match
            raise TileError('Mismatch with loaded analog state: '
                            'Hidden parameter structure is unexpected.')
        if not isinstance(weights, Tensor):
            weights = from_numpy(array(weights))
        self.tile.set_weights(weights)

        if not isinstance(hidden_parameters, Tensor):
            hidden_parameters = from_numpy(array(hidden_parameters))
        self.tile.set_hidden_parameters(hidden_parameters)

        self.tile.set_learning_rate(analog_lr)

        # re-generate shared weights (CPU)
        if shared_weights_if:
            if not hasattr(self, SN.SHARED_WEIGHTS):
                # this is needed when pkl loading
                self.shared_weights = shared_weights

            with no_grad():
                # always new will be populated with set weights.
                self.shared_weights.data = zeros(d_size, x_size, requires_grad=True)
            self.ensure_shared_weights()
        else:
            self.shared_weights = None

        # Regenerate context but keep the object ID
        if not hasattr(self, SN.CONTEXT):  # when loading
            self.analog_ctx = AnalogContext(self, parameter=analog_ctx)
        self.analog_ctx.reset(self)
        self.analog_ctx.set_data(analog_ctx.data)

        # set scales
        self.out_scaling_alpha = None
        self.mapping_scales = None
        self.init_mapping_scales()
        self.init_learned_out_scales()

        if self.out_scaling_alpha is None and learned_out_scales is not None:
            if mapping_scales is None:
                mapping_scales = 1.0
            x = learned_out_scales.view(learned_out_scales.numel()).clone()
            mapping_scales = mapping_scales * x
            learned_out_scales = None

        self.set_mapping_scales(mapping_scales)
        self.set_learned_out_scales(learned_out_scales)

        if alpha_scale is not None:
            # legacy. We apply the alpha scale instaed of the
            # out_scaling_alpha when loading. The alpha_scale
            # mechansim is now replaced with the out scaling factors
            #
            # Caution: will overwrite the loaded out_scaling_alphas
            # if they would exist also (should not be for old checkpoints)

            self.set_mapping_scales(alpha_scale)

        if to_device.type.startswith('cuda'):
            self.cuda(to_device)

        if alpha_scale is not None:
            # legacy. We apply the alpha scale instaed of the
            # out_scaling_alpha when loading. The alpha_scale
            # mechansim is now replaced with the out scaling factors
            #
            # Caution: will overwrite the loaded out_scaling_alphas
            # if they would exist also (should not be for old checkpoints)

            self.set_mapping_scales(alpha_scale)

    def _create_simulator_tile(
            self,
            x_size: int,
            d_size: int,
            rpu_config: RPUConfigGeneric
    ) -> Union[tiles.FloatingPointTile, tiles.AnalogTile]:
        """Create a simulator tile.

        Args:
            x_size: input size
            d_size: output size
            rpu_config: resistive processing unit configuration

        Returns:
            a simulator tile based on the specified configuration.
        """
        raise NotImplementedError

    def _combine_weights(self, weights: Union[Tensor, ArrayLike],
                         biases: Optional[Union[Tensor, ArrayLike]] = None) -> Tensor:
        """ Helper to combines weights and biases

        In any case, a detached cpu weight and bias copy will be returned.

        Args:
            weights: weights without the bias
            biases: The bias vector if available

        Returns:
            combined weights with biases

        Raises:
            ValueError: if the tile has bias but ``bias`` has not been
                specified.
        """
        if not isinstance(weights, Tensor):
            weights = from_numpy(array(weights))
        weights = weights.clone().detach().cpu().to(float32)

        if self.bias:
            # Create a ``[out_size, in_size (+ 1)]`` matrix.
            if biases is None:
                raise ValueError('Analog tile has a bias, but no bias given')

            if not isinstance(biases, Tensor):
                biases = from_numpy(array(biases))

            biases = unsqueeze(biases.clone().detach().cpu().to(float32), 1)
            return cat((weights, biases), dim=1)
        # Use only the ``[out_size, in_size]`` matrix.
        return weights

    def _separate_weights(self, combined_weights: Tensor) -> Tuple[Tensor, Optional[Tensor]]:
        """ Helper to separate the combined weights and biases
        """
        # Split the internal weights (and potentially biases) matrix.
        if self.bias:
            # combined_weights is [out_size, in_size (+ 1)].
            return Tensor(combined_weights[:, :-1]), Tensor(combined_weights[:, -1])

        return combined_weights, None

    def set_weights(
            self,
            weights: Tensor,
            biases: Optional[Tensor] = None,
            apply_weight_scaling: bool = False,
            weight_scaling_omega: Optional[float] = None
    ) -> None:
        """Set the tile weights (and biases).

        Sets the internal tile weights to the specified values, and also the
        internal tile biases if the tile was set to use bias (via
        ``self.bias``).

        Note:
           This setting is **not** hardware realistic. Use the
           :meth:`program_weights` for a realistic weight transfer.

        Args:
            weights: ``[out_size, in_size]`` weight matrix.
            biases: ``[out_size]`` bias vector. This parameter is required if
                ``self.bias`` is ``True``, and ignored otherwise.
            apply_weight_scaling: Whether to rescale the given weight matrix
                and populate the digital output scaling factors as
                specified in the configuration
                :class:`~aihwkit.configs.utils.MappingParameter`. A
                new ``weight_scaling_omega`` can be given. Note that
                this will overwrite the existing digital out scaling
                factors.
            weight_scaling_omega: The weight scaling omega factor (see
                :class:`~aihwkit.configs.utils.MappingParameter`). If
                given explicitly here, it will overwrite the value in
                the mapping field.

        Returns:
            None.
        """
        self.reference_combined_weights = None
        combined_weights = self._combine_weights(weights, biases)

        if apply_weight_scaling:
            combined_weights = self.apply_weight_scaling(combined_weights,
                                                         weight_scaling_omega)
        return self.tile.set_weights(combined_weights)

    def get_weights(self, apply_weight_scaling: bool = False
                    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the tile weights (and biases).

        Gets the tile weights and extracts the mathematical weight
        matrix and biases (if present, by determined by the ``self.bias``
        parameter).

        Note:
             The returned weight is a copy of the internal weights (not a
             pointer) and is always on CPU and detached.

        Note:
             This is **not** a hardware realistic weight readout. Use
            :meth:`read_weights` for a realistic transfer.

        Args:
            apply_weight_scaling: Whether to return the weights with the
                (digital) output scaling factors applied. Note the
                "logical" weights of the layer which the DNN is
                effectively using are those with the output scales
                applied. If ``apply_weight_scaling`` is set to False, then
                only the weight values that is programmed onto the
                crossbar array are returned, without applying the
                digital scales.

        Returns:
            a tuple where the first item is the ``[out_size, in_size]`` weight
            matrix; and the second item is either the ``[out_size]`` bias vector
            or ``None`` if the tile is set not to use bias.

        """
        # Retrieve the internal weights (and potentially biases) matrix.
        combined_weights = self.tile.get_weights()
        weights, biases = self._separate_weights(combined_weights)

        if not apply_weight_scaling:
            return weights, biases

        alpha = self.get_scales()
        if alpha is not None:
            alpha = alpha.detach().cpu()
            return weights * alpha.view(-1, 1), biases * alpha if self.bias else None
        return weights, biases

    @no_grad()
    def program_weights(self, from_reference: bool = True,
                        x_values: Optional[Tensor] = None,
                        learning_rate: float = 0.1,
                        max_iter: int = 10000,
                        tolerance: Optional[float] = 0.01,
                        w_init: Union[float, Tensor] = 0.01) -> None:
        """Programm the target weights into the conductances using the
        pulse update defined.

        Programming is done using the defined tile-update (e.g. SGD)
        and matching inputs (`x_values` by default `eye`).

        Args:

            from_reference: Whether to use weights from reference
                (those that were initally set with `set_weights`) or
                the current weights.
            x_values: Values to use for the read-and verify. If none
                are given, unit-vectors are used
            learning_rate: Learning rate of the optimization
            max_iter: max number of batches for the iterative programming
            tolerance: Stop the iteration loop early if the mean
                output deviation is below this number. Given in
                relation to the max output.
            w_init: initial weight matrix to start from. If given as
                float, weights are set uniform random in `[-w_init,
                w_init]`. This init weight is given directly in
                normalized conductance units and should include the
                bias row if existing.
        """

        if not from_reference or self.reference_combined_weights is None:
            self.reference_combined_weights = self.tile.get_weights()
        target_weights = self.reference_combined_weights

        if x_values is None:
            x_values = eye(self.tile.get_x_size())
        x_values = x_values.to(self.device)
        target_values = x_values @ target_weights.to(self.device).T

        target_max = target_values.abs().max().item()
        if isinstance(w_init, Tensor):
            self.tile.set_weights(w_init)
        else:
            self.tile.set_weights_uniform_random(-w_init, w_init)

        lr_save = self.tile.get_learning_rate()
        self.tile.set_learning_rate(learning_rate)

        for _ in range(max_iter):
            y = self.tile.forward(x_values, False)
            error = y - target_values
            if tolerance is not None and (error.abs().mean().item()
                                          / target_max) < tolerance:
                break
            self.tile.update(x_values, error, False)

        self.tile.set_learning_rate(lr_save)

    @no_grad()
    def read_weights(self,
                     apply_weight_scaling: bool = False,
                     x_values: Optional[Tensor] = None,
                     over_sampling: int = 10
                     ) -> Tuple[Tensor, Optional[Tensor]]:
        """Reads the weights (and biases) in a realistic manner
        by using the forward pass for weights readout.

        Gets the tile weights and extracts the mathematical weight
        matrix and biases (if present, by determined by the ``self.bias``
        parameter).

        The weight will not be directly read, but linearly estimated
        using random inputs using the analog forward pass.

        Note:

            If the tile includes digital periphery (e.g. out scaling),
            these will be applied. Thus this weight is the logical
            weights that correspond to the weights in an FP network.

        Note:
            weights are estimated using the ``lstsq`` solver from torch.

        Args:
            apply_weight_scaling: Whether to rescale the given weight matrix
                and populate the digital output scaling factors as
                specified in the configuration
                :class:`~aihwkit.configs.utils.MappingParameter`. A
                new ``weight_scaling_omega`` can be given. Note that
                this will overwrite the existing digital out scaling
                factors.

            x_values: Values to use for estimating the matrix. If
                not given, inputs are standard normal vectors.

            over_sampling: If ``x_values`` is not given,
                ``over_sampling * in_size`` random vectors are used
                for the estimation

        Returns:
            a tuple where the first item is the ``[out_size, in_size]`` weight
            matrix; and the second item is either the ``[out_size]`` bias vector
            or ``None`` if the tile is set not to use bias.

        """

        if x_values is None:
            x_values = randn(self.in_size * over_sampling, self.in_size,
                             device=self.device,
                             dtype=float32)
        else:
            x_values = x_values.to(self.device)

        y_values = self.forward(x_values, is_test=True)

        # calculate pseudo inverse (with bias, if necessary)
        if self.bias:
            ones_column = ones(self.in_size * over_sampling, 1,
                               device=self.device, dtype=float32)
            x_values = cat([x_values, ones_column], axis=1)

        est_weights = lstsq(x_values, y_values).solution.T.cpu()
        weights, biases = self._separate_weights(est_weights)

        if not apply_weight_scaling:
            # we de-apply (devide) because we want to use the full self.forward above
            alpha = self.get_scales()
            if alpha is not None:
                alpha = alpha.detach().cpu()
                return weights / alpha.view(-1, 1), biases / alpha if self.bias else None
        return weights, biases

    def apply_weight_scaling(
            self,
            combined_weights: Tensor,
            weight_scaling_omega: Optional[float] = None
    ) -> Tensor:
        r"""Set the tile weights (and biases) in a scaled fashion.

        Scales the weights by a layerwise scale or columnwise scale (if
        ``weight_scaling_columnwise`` is set), that is then applied in digital
        at the output of forward and backward pass, and the learning rate for
        this tile is adjusted accordingly.

        If layerwise scale is chosen, weights are scaled by
        :math:`\omega/\max_{ij} |w_{ij}|` and the global digital factor
        :math:`alpha` is set to :math:`\max_{ij} |w_{ij}|/\omega`.

        It can be shown that such a constant factor greatly improves the SNR and
        training accuracy as the full weight range of the analog devices are
        used. See also `Rasch, Gokmen & Haensch (2019)`_ for more details.

        Args:
            combined_weights: ``[d_size, x_size]`` weight matrix.
            weight_scaling_omega: where the weight max should be mapped in terms of
                the weight range. Note that for ``omega`` larger than
                the maximal weight of the device, weights will get
                clipped for most devices. If this parameter is not
                given, it will default to the ``weight_scaling_omega``
                value set in the
                :class:`~aihwkit.configs.utils.MappingParameter` of the
                ``rpu_config``

        Returns:
            scaled weights.


        .. _`Rasch, Gokmen & Haensch (2019)`: https://arxiv.org/abs/1906.02698

        """
        # Prepare the array expected by the pybind function, appending the
        # biases row if needed.
        if not hasattr(self.rpu_config, 'mapping'):
            return combined_weights

        mapping = self.rpu_config.mapping  # type: MappingParameter
        omega = weight_scaling_omega
        if omega is None:
            omega = mapping.weight_scaling_omega

        if omega is not None and omega > 0:
            # Apply the scaling
            if mapping.weight_scaling_columnwise:
                weight_max, _ = torch_max(abs(combined_weights), 1, keepdim=True)
            else:
                weight_max = torch_max(abs(combined_weights)).view(1)

            alpha = weight_max / omega

            alpha[alpha == 0.0] = 1.0
            combined_weights = combined_weights / alpha

            self.set_scales(alpha)
        return combined_weights

    @no_grad()
    def get_mapping_scales(self) -> Optional[Tensor]:
        """Get the scales used for the weight mapping.

        Returns:

           Mapping scales: the vector (or scalar) that is used to
           determine the mapping into (norm) conductance units. These
           scales are used at the output of the analog MVM.
        """
        return self.mapping_scales

    @no_grad()
    def set_mapping_scales(self, mapping_scales: Optional[Union[Tensor, float]]) -> None:
        """Set the scales used for the weight mapping.

        Args:
           mapping_scales: Vector (or scalar) used for the mapping
           of weights into conductance units. This mapping is never in
           the SGD graph but might get initialized when
           ``weight_scaling_omega`` is used or remapping is enforced.
        """

        if mapping_scales is None:
            self.mapping_scales = None
            return

        if isinstance(mapping_scales, float):
            if self.mapping_scales is None:
                self.mapping_scales = ones((1, ),
                                           dtype=float32,
                                           device=self.device,
                                           requires_grad=False)
            self.mapping_scales[:] = mapping_scales
            return

        if isinstance(self.mapping_scales, Tensor) and len(mapping_scales) == 1:
            self.mapping_scales[:] = mapping_scales.to(self.device)
            return

        self.mapping_scales = mapping_scales.flatten().to(self.device)

    @no_grad()
    def init_mapping_scales(self) -> None:
        """Helper function to initialize the mapping scales used to scale the
        weights in digital and determine the conductance conversion.

        Note:
            This method is called from the constructor.
        """
        if not hasattr(self.rpu_config, 'mapping'):
            self.set_mapping_scales(None)
            return

        mapping = self.rpu_config.mapping  # type: MappingParameter
        mapping_scales = None
        if mapping.weight_scaling_omega:
            if mapping.weight_scaling_columnwise:
                mapping_scales = ones((self.out_size, ),
                                      dtype=float32,
                                      device=self.device,
                                      requires_grad=False)
            else:
                mapping_scales = ones((1, ),
                                      dtype=float32,
                                      device=self.device,
                                      requires_grad=False)
        self.set_mapping_scales(mapping_scales)

    @no_grad()
    def init_input_processing(self) -> None:
        """Helper function to initialize the input processing.

        Note:
            This method is called from the constructor.

        Raises: ConfigError in case ``manage_output_clipping`` is
            enabled but not supported.
        """
        self.input_range = None
        if not hasattr(self.rpu_config, 'pre_post'):
            return

        ir_params = self.rpu_config.pre_post.input_range  # type: InputRangeParameter
        if ir_params.enable:
            self.input_range_update_idx = 0
            self.input_range = full((1,), ir_params.init_value, dtype=float32,
                                    device=self.device, requires_grad=True)

            if (ir_params.manage_output_clipping
                    and not ir_params.supports_manage_output_clipping(self.rpu_config)):
                raise ConfigError("RPU Config does not support `manage_output_clipping`.")

    @no_grad()
    def set_scales(self, scales: Union[Tensor, float]) -> None:
        """Set all scales with a new scale.

        This will set the mapping scales to ``scales`` and set all other scales to 1.

        Args:
            scales: scales to set.
        """

        self.set_mapping_scales(scales)
        self.set_learned_out_scales(1.0)

    @no_grad()
    def get_scales(self) -> Optional[Tensor]:
        """ Set all scales with a new scale.

        Returns:
            Scale tensor if any scale exist else None.
        """

        learned_out_scales = self.get_learned_out_scales()
        mapping_scales = self.get_mapping_scales()
        if mapping_scales is None and learned_out_scales is None:
            return None
        if mapping_scales is None:
            return learned_out_scales
        if learned_out_scales is None:
            return mapping_scales
        return mapping_scales * learned_out_scales

    @no_grad()
    def get_learned_out_scales(self) -> Tensor:
        """Get the learned_out_scaled that can be used add an output scale to
        the weights, that is learned.

        Returns:
            tensor: learned_out_scales

        """
        return self.out_scaling_alpha

    @no_grad()
    def init_learned_out_scales(self) -> None:
        """Helper function to initialize the learned out scaling used to scale the
        weights in digital.

        Note:
            This method is called from the constructor.
        """

        if not hasattr(self.rpu_config, 'mapping'):
            return

        mapping = self.rpu_config.mapping  # type: ignore
        if mapping.learn_out_scaling:
            if mapping.out_scaling_columnwise:
                self.out_scaling_alpha = ones((self.out_size, ),
                                              dtype=float32,
                                              device=self.device,
                                              requires_grad=True)
            else:
                self.out_scaling_alpha = ones((1, ),
                                              dtype=float32,
                                              device=self.device,
                                              requires_grad=True)

    @no_grad()
    def set_learned_out_scales(self, alpha: Union[Tensor, float]) -> None:
        """Helper function to set the out scaling alpha used to scale the
        weights in digital.

        Note:
            Will be a no-op in case :meth:`~init_learned_out_scales`
            was not called

        Caution:
            Will not check the correct size of the given alpha.

        Args:
            alpha: out scales as a parameter that is learned.

        """
        if self.out_scaling_alpha is None:
            return

        if isinstance(self.out_scaling_alpha, Parameter):
            self.out_scaling_alpha.data[:] = squeeze(as_tensor(alpha)).to(self.device)
        elif isinstance(self.out_scaling_alpha, Tensor):
            self.out_scaling_alpha[:] = squeeze(as_tensor(alpha)).to(self.device)
        else:
            self.out_scaling_alpha = squeeze(as_tensor(alpha)).to(self.device)

    def apply_out_scaling(self, values: Tensor,
                          tensor_view: Optional[Tuple[int, ...]] = None) -> Tensor:
        """Apply the learned out scaling to the given tensor.

        Args:
            values: tensor to apply scaling to.
            tensor_view: view to cast the out scalings before multiplication

        Returns:
            output tensor with applied out scaling factors
        """
        if self.out_scaling_alpha is not None:
            if tensor_view is None:
                tensor_view = self._get_tensor_view(values.dim(),
                                                    0 if self.out_trans else values.dim() - 1)
            return values * self.out_scaling_alpha.view(*tensor_view)
        return values

    @no_grad()
    def apply_input_range(self, values: Tensor, update_from_data: bool = False) -> Tensor:
        """ Apply the input clipping.

        Args:
            values: tensor to clip
            update_from_data: whether to update from data if applicable

        Returns:
            clipped output tensor
        """
        if self.input_range is None:
            return values

        if update_from_data:
            ir_params = self.rpu_config.pre_post.input_range  # type: ignore
            idx = self.input_range_update_idx
            if idx < ir_params.init_from_data:
                self.input_range.data = (self.input_range.data * idx
                                         + ir_params.init_std_alpha * values.std()
                                         ) / (idx + 1)
                self.input_range_update_idx += 1

        self.input_range.data = self.input_range.data.abs()
        return clamp(values,
                     min=-self.input_range.item(),  # pylint: disable=invalid-unary-operand-type
                     max=self.input_range.item())

    def set_learning_rate(self, learning_rate: float) -> None:
        """Set the tile learning rate.

        Set the tile learning rate to ``-learning_rate``. Note that the
        learning rate is always taken to be negative (because of the meaning in
        gradient descent) and positive learning rates are not supported.

        Args:
            learning_rate: the desired learning rate.

        Returns:
            None.
        """
        return self.tile.set_learning_rate(learning_rate)

    def get_learning_rate(self) -> float:
        """Return the tile learning rate.

        Returns:
            float: the tile learning rate.
        """
        return self.tile.get_learning_rate()

    @no_grad()
    def decay_weights(self, alpha: float = 1.0) -> None:
        """Decays the weights once according to the decay parameters of the tile.

        Args:
            alpha: additional decay scale (such as LR). The base decay
                rate is set during tile init.

        Returns:
            None.
        """
        return self.tile.decay_weights(alpha)

    @no_grad()
    def drift_weights(self, delta_t: float = 1.0) -> None:
        """Drifts the weights once according to the drift parameters of the
        tile.

        See also :class:`~aihwkit.simulator.configs.utils.DriftParameter`.

        Args:
            delta_t: Time since last drift call.

        Returns:
            None.
        """
        return self.tile.drift_weights(delta_t)

    @no_grad()
    def diffuse_weights(self) -> None:
        """Diffuses the weights once according to the diffusion parameters of
        the tile.

        The base diffusion rate is set during tile init.

        Returns:
            None
        """
        return self.tile.diffuse_weights()

    @no_grad()
    def reset_columns(
            self,
            start_column_idx: int = 0,
            num_columns: int = 1,
            reset_prob: float = 1.0
    ) -> None:
        r"""Reset (a number of) columns according to the reset parameters of the tile.

        Resets the weights with device-to-device and cycle-to-cycle
        variability (depending on device type), typically:

        .. math::
            W_{ij} = \xi*\sigma_\text{reset} + b^\text{reset}_{ij}

        The reset parameters are set during tile init.

        Args:
            start_column_idx: a start index of columns (0..x_size-1)
            num_columns: how many consecutive columns to reset (with circular warping)
            reset_prob: individual probability of reset.

        Returns:
            None
        """
        return self.tile.reset_columns(start_column_idx, num_columns, reset_prob)

    @no_grad()
    def reset(
            self,
            reset_prob: float = 1.0
    ) -> None:
        r"""Reset the updated device tile according to the reset parameters of the tile.

        Resets the weights with device-to-device and cycle-to-cycle
        variability (depending on device type), typically:

        .. math::
            W_{ij} = \xi*\sigma_\text{reset} + b^\text{reset}_{ij}

        The reset parameters are set during tile init.

        Args:
            reset_prob: individual probability of reset.

        Returns:
            None
        """
        return self.tile.reset_columns(0, -1, reset_prob)

    def cpu(self) -> 'BaseTile':
        """Return a copy of this tile in CPU memory.

        Returns:
            self in case of CPU

        """
        if not self.is_cuda:
            return self

        state_dict = self.__getstate__()
        for value in state_dict.values():
            if isinstance(value, AnalogContext):
                value.data = value.data.cpu()
        self.__setstate__(state_dict)
        return self

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory."""
        raise NotImplementedError

    def _get_tensor_view(self, ndim: int, dim: int) -> tuple:
        """Return the tensor view for ndim vector at dim.

        Args:
            ndim: number of dimensions
            dim: the dimension to set to -1

        Returns:
            List of ones with the `dim`` index sets to -1
        """
        tensor_view = [1] * ndim
        tensor_view[dim] = -1
        return tuple(tensor_view)

    @no_grad()
    def pre_forward(self, x_input: Tensor, dim: int,
                    is_test: bool = False, ctx: Any = None) -> Tensor:
        """Operations before the actual forward step for pre processing.

        By default, this is an no-op. However, it could be overridden
        in derived tile classes.

        Args:
            x_input: input tensor for the analog MVM of the tile.
            dim: input channel dimension, ie the x_size dimension
            is_test: whether in eval mode
            ctx: torch auto-grad context [Optional]

        Returns:
            Output tensor of the same shape
        """
        # pylint: disable=unused-argument
        if self.input_range is not None:
            x_input = self.apply_input_range(x_input, not is_test) / self.input_range
        return x_input

    @no_grad()
    def post_forward(self, x_output: Tensor,
                     dim: int,
                     is_test: bool = False,
                     ctx: Any = None) -> Tensor:
        """Operations after the actual forward step for post processing.

        Args:
            x_output:  tensor that is the output from the forward pass of the tile
            dim: output channel dimension, ie the d_size dimension
            is_test: whether in eval mode
            ctx: torch auto-grad context [Optional]

        Returns:
            Output tensor of the same shape
        """
        # pylint: disable=unused-argument
        scale = None
        if self.input_range is not None:
            scale = self.input_range

            # if output clip determines input clip learning
            ir_params = self.rpu_config.pre_post.input_range  # type: ignore
            if ctx is not None and ir_params.manage_output_clipping:
                out_bound = self.rpu_config.forward.out_bound * 0.999  # type: ignore
                output_percentage = (x_output.abs() < out_bound).float().mean()
                ctx.output_percentage = output_percentage

        if self.mapping_scales is not None:
            tensor_view = self._get_tensor_view(x_output.dim(), dim)
            if scale is not None:
                scale = scale * self.get_mapping_scales().view(tensor_view)
            else:
                scale = self.get_mapping_scales().view(tensor_view)

        if scale is not None:
            return x_output * scale
        return x_output

    @no_grad()
    def forward(self, x_input: Tensor, is_test: bool = False, ctx: Any = None) -> Tensor:
        """Perform the forward pass.

        Calls first the ``pre_forward``, then the tile forward, and
        finally the ``post_forward`` step.

        Note:

            The full forward pass is not using autograd, thus all pre
            and post functions need to be handled appropriately in the
            pre/post backward functions.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            is_test: whether to assume testing mode.
            ctx: torch auto-grad context [Optional]

        Returns:
            torch.Tensor: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        """
        # We use no-grad as we do it explicitly in the optimizer.
        x_input = self.pre_forward(x_input,
                                   0 if self.in_trans else x_input.dim() - 1,
                                   is_test, ctx)
        x_output = self.tile.forward(x_input, self.bias, self.in_trans,
                                     self.out_trans, is_test, self.non_blocking)
        return self.post_forward(x_output,
                                 0 if self.out_trans else x_output.dim() - 1,
                                 is_test, ctx)

    @no_grad()
    def pre_backward(self, d_input: Tensor, dim: int, ctx: Any = None) -> Tensor:
        """Operations before the actual backward step for pre processing.

        By default, this is an no-op. However, it could be overridden
        in derived tile classes.

        Args:
            d_input: The input tensor from to the analog MVM of the tile.
            dim: the dim of the d_size dimension
            ctx: torch auto-grad context [Optional]

        Returns:
            The preprocessed tensor of the same shape
        """
        # pylint: disable=unused-argument
        if self.mapping_scales is not None:
            tensor_view = self._get_tensor_view(d_input.dim(), dim)
            return d_input * self.get_mapping_scales().view(tensor_view)
        return d_input

    @no_grad()
    def post_backward(self, d_output: Tensor, dim: int, ctx: Any = None) -> Tensor:
        """Operations after the actual backward step for post processing.

        Here, the mapping scales are applied if exist.

        Args:
            d_output: The output tensor from the analog MVM of the tile.
            dim: the dim of the x_size dimension
            ctx: torch auto-grad context [Optional]

        Returns:
            The postprocessed tensor of the same shape
        """
        # pylint: disable=unused-argument

        if self.input_range is not None and ctx is not None:
            # compute gradient of the clip
            x_input,  = ctx.saved_tensors
            ir_params = self.rpu_config.pre_post.input_range  # type: ignore

            upper_thres = x_input >= self.input_range
            lower_thres = x_input <= -self.input_range  # pylint: disable=invalid-unary-operand-type

            grad = zeros_like(self.input_range)

            grad += clamp(upper_thres * d_output, min=None, max=0.0).sum()
            grad -= clamp(lower_thres * d_output, min=0.0, max=None).sum()

            if ir_params.gradient_relative:
                grad *= self.input_range
            grad *= ir_params.gradient_scale

            if ir_params.manage_output_clipping:
                output_percentage = getattr(ctx, 'output_percentage', 1.0)
                grad -= (1.0 - output_percentage) * self.input_range * (
                    output_percentage < ir_params.output_min_percentage)

            if ir_params.decay > 0:
                percentage = (x_input.abs() < self.input_range).float().mean()
                grad += ir_params.decay * self.input_range * (
                    percentage > ir_params.input_min_percentage)

            if self.input_range.grad is None:
                self.input_range.grad = grad
            else:
                self.input_range.grad += grad

        return d_output

    @no_grad()
    def backward(self, d_input: Tensor, ctx: Any = None) -> Tensor:
        """Perform the backward pass.

        Args:
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.
            ctx: torch auto-grad context [Optional]

        Returns:
            torch.Tensor: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
        """
        d_input = self.pre_backward(d_input, 0 if self.out_trans else d_input.dim() - 1,
                                    ctx)
        d_output = self.tile.backward(d_input, self.bias, self.out_trans, self.in_trans,
                                      self.non_blocking)
        return self.post_backward(d_output, 0 if self.in_trans else d_output.dim() - 1,
                                  ctx)

    @no_grad()
    def pre_update(self, x_input: Tensor, x_dim: int,
                   d_input: Tensor, d_dim: int) -> Tuple[Tensor, Tensor]:
        """Operations before the actual update step for pre processing.

        Be default, if the mapping scales are used, the ``d_input``
        will be divided by the mapping scales to compensate for the
        conductance mapping.

        Caution:

            The ``x_input`` and ``d_input`` here are the *original* inputs
            to the ``forward` and ``backward`` methods, thus the
            ``pre_forward`` and ``pre_backward`` function are *not*
            applied, and might need to be applied again here.

        Args:
            x_input: The forward input tensor.
            x_dim: the dim of the x_size dimension of the forward input.
            d_input: The backward (gradient) input tensor.
            d_dim: the dim of the d_size dimension of the backward input.

        Returns:
           Tuple of the preprocessed x_input and d_input tensors of the same shape

        """
        # pylint: disable=unused-argument
        if self.input_range is not None:
            x_input = self.apply_input_range(x_input, False) / self.input_range

        if self.mapping_scales is not None:
            tensor_view = self._get_tensor_view(d_input.dim(), d_dim)
            return x_input, d_input / self.get_mapping_scales().view(tensor_view)

        return x_input, d_input

    @no_grad()
    def update(self, x_input: Tensor, d_input: Tensor) -> None:
        """Perform the update pass.

        Calls the ``pre_update`` method to pre-process the inputs.

        Args:
            x_input: ``[..., in_size]`` tensor. If ``in_trans`` is set, ``[in_size, ...]``.
            d_input: ``[..., out_size]`` tensor. If ``out_trans`` is set, ``[out_size, ...]``.

        Returns:
            None
        """
        x_input, d_input = self.pre_update(x_input,
                                           0 if self.in_trans else x_input.dim() - 1,
                                           d_input,
                                           0 if self.out_trans else d_input.dim() - 1)
        return self.tile.update(x_input, d_input, self.bias,
                                self.in_trans, self.out_trans, self.non_blocking)

    def get_hidden_parameters(self) -> OrderedDict:
        """Get the hidden parameters of the tile.

        Returns:
            Ordered dictionary of hidden parameter tensors.
        """
        names = self.tile.get_hidden_parameter_names()
        hidden_parameters = self.tile.get_hidden_parameters().detach_()

        ordered_parameters = OrderedDict()
        for idx, name in enumerate(names):
            ordered_parameters[name] = hidden_parameters[idx].clone()

        return ordered_parameters

    def set_hidden_parameters(self, ordered_parameters: OrderedDict) -> None:
        """Set the hidden parameters of the tile.

        Caution:
            Usually the hidden parameters are drawn according to the
            parameter definitions (those given in the RPU config). If
            the hidden parameters are arbitrary set by the user, then
            this correspondence might be broken. This might cause problems
            in the learning, in particular, the `weight granularity`
            (usually ``dw_min``, depending on the device) is needed for
            the dynamic adjustment of the bit length
            (``update_bl_management``, see
            :class:`~aihwkit.simulator.configs.utils.UpdateParameters`).

            Currently, the new ``dw_min`` parameter is tried to be
            estimated from the average of hidden parameters if the
            discrepancy with the ``dw_min`` from the definition is too
            large.

        Args:
            ordered_parameters: Ordered dictionary of hidden parameter tensors.

        Raises:
            TileError: In case the ordered dict keys do not conform
                with the current rpu config tile structure of the hidden
                parameters
        """
        if len(ordered_parameters) == 0:
            return

        hidden_parameters = stack(list(ordered_parameters.values()), dim=0)
        names = self.tile.get_hidden_parameter_names()
        if names != list(ordered_parameters.keys()):
            raise TileError('Mismatch with loaded analog state:'
                            'Hidden parameter structure is unexpected.')

        self.tile.set_hidden_parameters(hidden_parameters)

    def get_hidden_update_index(self) -> int:
        """Get the current updated device index of the hidden devices.

        Usually this is 0 as only one device is present per
        cross-point for many tile RPU configs. However, some RPU
        configs maintain internally multiple devices per cross-point
        (e.g. :class:`~aihwkit.simulator.config.devices.VectorUnitCell`).

        Returns:
            The next mini-batch updated device index.

        Note:
            Depending on the update and learning policy implemented
            in the tile, updated devices might switch internally as
            well.
        """
        return self.tile.get_hidden_update_index()

    def set_hidden_update_index(self, index: int) -> None:
        """Set the current updated hidden device index.

        Usually this is ignored and fixed to 0 as only one device is
        present per cross-point. Other devices, might not allow
        explicit setting as it would interfere with the implemented
        learning rule. However, some tiles have internally
        multiple devices per cross-point (eg. unit cell) that can be
        chosen depending on the update policy.

        Args:
            index: device index to be updated in the next mini-batch

        Note:
            Depending on the update and learning policy implemented
            in the tile, updated devices might switch internally as
            well.
        """
        self.tile.set_hidden_update_index(index)

    def is_indexed(self) -> bool:
        """Returns whether index matrix for convolutions has been set.

        Returns:
           Whether index matrix has been set
        """
        return self.tile.has_matrix_indices()

    def set_indexed(self, indices: Tensor, image_sizes: List) -> None:
        """Set the index matrix for convolutions and switches to
        indexed forward/backward/update versions.

        Args:
            indices : torch.tensor with int indices
            image_sizes: [C_in, H_in, W_in, H_out, W_out] sizes

        Raises:
            ValueError: if ``image_sizes`` does not have valid dimensions.
            TileError: if the tile uses transposition.
        """
        if len(image_sizes) not in (3, 5, 7):
            raise ValueError('image_sizes expects 3, 5 or 7 sizes '
                             '[C_in, (D_in), H_in, (W_in), (D_out), H_out, (W_out)]')

        if self.in_trans or self.out_trans:
            raise TileError('Transposed indexed versions not supported (assumes NC(D)HW)')

        self.image_sizes = image_sizes
        self.tile.set_matrix_indices(indices)

    @no_grad()
    def forward_indexed(self, x_input: Tensor, is_test: bool = False,
                        ctx: Any = None) -> Tensor:
        """Perform the forward pass for convolutions.

        Depending on the input tensor size it performs the forward pass for a
        2D image or a 3D one.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            is_test: whether to assume testing mode.
            ctx: torch auto-grad context [Optional]

        Returns:
            torch.Tensor: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Raises:
            TileError: if the indexed tile has not been initialized, or if
                ``self.images_sizes`` does not have a valid dimennion.
        """
        if not self.image_sizes:
            raise TileError('self.image_sizes is not initialized. Please use '
                            'set_indexed()')

        n_batch = x_input.size(0)
        channel_out = self.out_size

        if len(self.image_sizes) == 3:
            _, _, height_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, height_out))
        elif len(self.image_sizes) == 5:
            _, _, _, height_out, width_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, height_out, width_out))
        elif len(self.image_sizes) == 7:
            _, _, _, _, depth_out, height_out, width_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, depth_out, height_out, width_out))
        else:
            raise TileError('self.image_sizes length is not 3, 5 or 7')

        x_input = self.pre_forward(x_input, 1, is_test, ctx)
        x_output = self.tile.forward_indexed(x_input, d_tensor, is_test, self.non_blocking)
        return self.post_forward(x_output, 1, is_test, ctx)

    @no_grad()
    def backward_indexed(self, d_input: Tensor, ctx: Any = None) -> Tensor:
        """Perform the backward pass for convolutions.

        Depending on the input tensor size it performs the backward pass for a
        2D image or a 3D one.

        Args:
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.
            ctx: torch auto-grad context [Optional]

        Returns:
            torch.Tensor: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.

        Raises:
            TileError: if the indexed tile has not been initialized, or if
                ``self.images_sizes`` does not have a valid dimennion.
        """
        if not self.image_sizes:
            raise TileError('self.image_sizes is not initialized. Please use '
                            'set_indexed()')

        n_batch = d_input.size(0)

        if len(self.image_sizes) == 3:
            channel_in, height_in, _ = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, height_in))
        elif len(self.image_sizes) == 5:
            channel_in, height_in, width_in, _, _ = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, height_in, width_in))
        elif len(self.image_sizes) == 7:
            channel_in, depth_in, height_in, width_in, _, _, _ \
                = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, depth_in, height_in, width_in))
        else:
            raise TileError('self.image_sizes length is not 3, 5 or 7')

        d_input = self.pre_backward(d_input, 1, ctx)
        d_output = self.tile.backward_indexed(d_input, x_tensor, self.non_blocking)
        return self.post_backward(d_output, 1, ctx)

    @no_grad()
    def update_indexed(self, x_input: Tensor, d_input: Tensor) -> None:
        """Perform the update pass for convolutions.

        Calls the ``pre_update`` methods to pre-process the inputs.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Returns:
            None
        """
        x_input, d_input = self.pre_update(x_input, 1, d_input, 1)
        return self.tile.update_indexed(x_input, d_input, self.non_blocking)

    @no_grad()
    def post_update_step(self) -> None:
        """Operators that need to be called once per mini-batch.

        Note:
            This function is called by the analog optimizer.

        Caution:

            If no analog optimizer is used, the post update steps will
            not be performed.
        """
        if self.rpu_config.device.requires_diffusion():  # type: ignore
            self.diffuse_weights()
        if self.rpu_config.device.requires_decay():  # type: ignore
            self.decay_weights()



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Analog RNN layers """

from typing import Any, List, Tuple, Type, Union
from torch import Tensor, stack, jit, cat
from torch.nn import ModuleList
from aihwkit.nn.modules.container import AnalogSequential


class AnalogRNNLayer(AnalogSequential):
    """Analog RNN Layer.

    Args:
        cell: RNNCell type (AnalogLSTMCell/AnalogGRUCell/AnalogVanillaRNNCell/
              AnalogLSTMCellSingleRPU)
        cell_args: arguments to RNNCell (e.g. input_size, hidden_size, rpu_configs)
    """
    # pylint: disable=abstract-method

    def __init__(self, cell: Type, *cell_args: Any):
        super().__init__()
        self.cell = cell(*cell_args)

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        return self.cell.get_zero_state(batch_size)

    def forward(
            self, input_: Tensor,
            state: Union[Tuple[Tensor, Tensor], Tensor]
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        # pylint: disable=arguments-differ
        inputs = input_.unbind(0)
        outputs = jit.annotate(List[Tensor], [])
        for input_item in inputs:
            out, state = self.cell(input_item, state)
            outputs += [out]
        return stack(outputs), state


class AnalogReverseRNNLayer(AnalogSequential):
    """ Analog RNN layer for direction.

    Args:
        cell: RNNCell type (AnalogLSTMCell/AnalogGRUCell/AnalogVanillaRNNCell)
        cell_args: arguments to RNNCell (e.g. input_size, hidden_size, rpu_configs)
    """
    def __init__(self, cell: Type, *cell_args: Any):
        super().__init__()
        self.cell = cell(*cell_args)

    @staticmethod
    def reverse(lst: List[Tensor]) -> List[Tensor]:
        """ Reverses the list of input tensors. """
        return lst[::-1]

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        return self.cell.get_zero_state(batch_size)

    def forward(self, input_: Tensor,
                state: Union[Tuple[Tensor, Tensor], Tensor]
                ) -> Tuple[Tensor, Union[Tuple[Tensor, Tensor], Tensor]]:
        # pylint: disable=arguments-differ
        inputs = self.reverse(input_.unbind(0))
        outputs = jit.annotate(List[Tensor], [])
        for input_values in inputs:
            out, state = self.cell(input_values, state)
            outputs += [out]
        return stack(self.reverse(outputs)), state


class AnalogBidirRNNLayer(AnalogSequential):
    """ Bi-directional analog RNN layer.

    Args:
        cell: RNNCell type (AnalogLSTMCell/AnalogGRUCell/AnalogVanillaRNNCell)
        cell_args: arguments to RNNCell (e.g. input_size, hidden_size, rpu_configs)
    """

    __constants__ = ['directions']

    def __init__(self, cell: Type, *cell_args: Any):
        super().__init__()

        self.directions = ModuleList([
            AnalogRNNLayer(cell, *cell_args),
            AnalogReverseRNNLayer(cell, *cell_args),
        ])

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        return [self.directions[0].get_zero_state(batch_size),
                self.directions[1].get_zero_state(batch_size)]

    def forward(self, input_: Tensor,
                states: List[Union[Tuple[Tensor, Tensor], Tensor]]
                ) -> Tuple[Tensor, List[Union[Tuple[Tensor, Tensor], Tensor]]]:
        # pylint: disable=arguments-differ
        # List[RNNState]: [forward RNNState, backward RNNState]
        outputs = jit.annotate(List[Tensor], [])
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])

        for direction, state in zip(self.directions, states):
            out, out_state = direction(input_, state)
            outputs += [out]
            output_states += [out_state]

        return cat(outputs, -1), output_states



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Analog cells for RNNs. """

from typing import Optional, Tuple
from collections import namedtuple

from torch import Tensor, sigmoid, tanh, zeros, cat

from aihwkit.nn.modules.container import AnalogSequential

from aihwkit.simulator.configs import InferenceRPUConfig
from aihwkit.nn.modules.base import RPUConfigAlias

LSTMState = namedtuple('LSTMState', ['hx', 'cx'])


class AnalogVanillaRNNCell(AnalogSequential):
    """Analog Vanilla RNN Cell.

    Args:
        input_size: in_features size for W_ih matrix
        hidden_size: in_features and out_features size for W_hh matrix
        bias: whether to use a bias row on the analog tile or not
        rpu_config: configuration for an analog resistive processing unit
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
    """
    # pylint: disable=abstract-method
    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ):
        super().__init__()

        # Default to InferenceRPUConfig
        if not rpu_config:
            rpu_config = InferenceRPUConfig()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = rpu_config.get_linear()(input_size, hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)
        self.weight_hh = rpu_config.get_linear()(hidden_size, hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        device = self.weight_ih.get_analog_tile_devices()[0]
        return zeros(batch_size, self.hidden_size, device=device)

    def forward(
            self,
            input_: Tensor,
            state: Tensor
    ) -> Tuple[Tensor, Tensor]:
        # pylint: disable=arguments-differ
        igates = self.weight_ih(input_)
        hgates = self.weight_hh(state)

        out = tanh(igates + hgates)

        return out, out  # output will also be hidden state


class AnalogLSTMCell(AnalogSequential):
    """Analog LSTM Cell.

    Args:
        input_size: in_features size for W_ih matrix
        hidden_size: in_features and out_features size for W_hh matrix
        bias: whether to use a bias row on the analog tile or not
        rpu_config: configuration for an analog resistive processing unit
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ):
        super().__init__()

        # Default to InferenceRPUConfig
        if not rpu_config:
            rpu_config = InferenceRPUConfig()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = rpu_config.get_linear()(input_size, 4 * hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)
        self.weight_hh = rpu_config.get_linear()(hidden_size, 4 * hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        device = self.weight_ih.get_analog_tile_devices()[0]
        return LSTMState(zeros(batch_size, self.hidden_size, device=device),
                         zeros(batch_size, self.hidden_size, device=device))

    def forward(self, input_: Tensor,
                state: Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:

        # pylint: disable=arguments-differ
        h_x, c_x = state
        gates = self.weight_ih(input_) + self.weight_hh(h_x)
        in_gate, forget_gate, cell_gate, out_gate = gates.chunk(4, 1)

        in_gate = sigmoid(in_gate)
        forget_gate = sigmoid(forget_gate)
        cell_gate = tanh(cell_gate)
        out_gate = sigmoid(out_gate)

        c_y = (forget_gate * c_x) + (in_gate * cell_gate)
        h_y = out_gate * tanh(c_y)

        return h_y, (h_y, c_y)


class AnalogLSTMCellCombinedWeight(AnalogSequential):
    """Analog LSTM Cell that use a combined weight for storing gates and inputs.

    Args:
        input_size: The number of expected features in the input `x`
        hidden_size: The number of features in the hidden state `h`
        bias: whether to use a bias row on the analog tile or not.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and during reading of the weights.
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ):
        super().__init__()

        # Default to InferenceRPUConfig
        if not rpu_config:
            rpu_config = InferenceRPUConfig()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight = rpu_config.get_linear()(input_size + hidden_size,
                                              4 * hidden_size,
                                              bias=bias,
                                              rpu_config=rpu_config,
                                              realistic_read_write=realistic_read_write)

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        device = self.weight.get_analog_tile_devices()[0]
        return LSTMState(zeros(batch_size, self.hidden_size, device=device),
                         zeros(batch_size, self.hidden_size, device=device))

    def forward(self, input_: Tensor,
                state: Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:

        # pylint: disable=arguments-differ
        h_x, c_x = state
        x_input = cat((input_, h_x), 1)
        gates = self.weight(x_input)
        in_gate, forget_gate, cell_gate, out_gate = gates.chunk(4, 1)

        in_gate = sigmoid(in_gate)
        forget_gate = sigmoid(forget_gate)
        cell_gate = tanh(cell_gate)
        out_gate = sigmoid(out_gate)

        c_y = (forget_gate * c_x) + (in_gate * cell_gate)
        h_y = out_gate * tanh(c_y)

        return h_y, (h_y, c_y)


class AnalogGRUCell(AnalogSequential):
    """Analog GRU Cell.

    Args:
        input_size: in_features size for W_ih matrix
        hidden_size: in_features and out_features size for W_hh matrix
        bias: whether to use a bias row on the analog tile or not
        rpu_config: configuration for an analog resistive processing unit
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ):
        super().__init__()

        # Default to InferenceRPUConfig
        if not rpu_config:
            rpu_config = InferenceRPUConfig()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = rpu_config.get_linear()(input_size, 3 * hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)
        self.weight_hh = rpu_config.get_linear()(hidden_size, 3 * hidden_size, bias=bias,
                                                 rpu_config=rpu_config,
                                                 realistic_read_write=realistic_read_write)

    def get_zero_state(self, batch_size: int) -> Tensor:
        """Returns a zeroed state.

        Args:
            batch_size: batch size of the input

        Returns:
           Zeroed state tensor
        """
        device = self.weight_ih.get_analog_tile_devices()[0]
        return zeros(batch_size, self.hidden_size, device=device)

    def forward(self, input_: Tensor, state: Tensor) -> Tuple[Tensor, Tensor]:

        # pylint: disable=arguments-differ

        g_i = self.weight_ih(input_)
        g_h = self.weight_hh(state)
        i_r, i_i, i_n = g_i.chunk(3, 1)
        h_r, h_i, h_n = g_h.chunk(3, 1)

        reset_gate = sigmoid(i_r + h_r)
        input_gate = sigmoid(i_i + h_i)
        new_gate = tanh(i_n + reset_gate * h_n)
        hidden_y = new_gate + input_gate * (state - new_gate)

        return hidden_y, hidden_y  # output will also be hidden state



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: i_common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import i_onnx_common_pb2 as i__onnx__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ei_common.proto\x12\x11\x61ihwx.inferencing\x1a\x13i_onnx_common.proto\"R\n\x0eOptimizerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"R\n\x0ePrimitiveProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"R\n\x0eTransformProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"\x85\x01\n\nLayerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\x12\x35\n\nstate_dict\x18\x03 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"[\n\x17\x41\x63tivationFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"\x9e\x01\n\x19LayerOrActivationFunction\x12I\n\x13\x61\x63tivation_function\x18\x01 \x01(\x0b\x32*.aihwx.inferencing.ActivationFunctionProtoH\x00\x12.\n\x05layer\x18\x02 \x01(\x0b\x32\x1d.aihwx.inferencing.LayerProtoH\x00\x42\x06\n\x04item\"R\n\x0eSchedulerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"U\n\x11LossFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12\x34\n\targuments\x18\x02 \x03(\x0b\x32!.aihwx.inferencing.AttributeProto\"c\n\x07Network\x12\x1a\n\x12weight_template_id\x18\x01 \x02(\t\x12<\n\x06layers\x18\x02 \x03(\x0b\x32,.aihwx.inferencing.LayerOrActivationFunction\"(\n\x07Version\x12\x0e\n\x06schema\x18\x01 \x02(\x03\x12\r\n\x05opset\x18\x02 \x02(\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'i_common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _OPTIMIZERPROTO._serialized_start=58
  _OPTIMIZERPROTO._serialized_end=140
  _PRIMITIVEPROTO._serialized_start=142
  _PRIMITIVEPROTO._serialized_end=224
  _TRANSFORMPROTO._serialized_start=226
  _TRANSFORMPROTO._serialized_end=308
  _LAYERPROTO._serialized_start=311
  _LAYERPROTO._serialized_end=444
  _ACTIVATIONFUNCTIONPROTO._serialized_start=446
  _ACTIVATIONFUNCTIONPROTO._serialized_end=537
  _LAYERORACTIVATIONFUNCTION._serialized_start=540
  _LAYERORACTIVATIONFUNCTION._serialized_end=698
  _SCHEDULERPROTO._serialized_start=700
  _SCHEDULERPROTO._serialized_end=782
  _LOSSFUNCTIONPROTO._serialized_start=784
  _LOSSFUNCTIONPROTO._serialized_end=869
  _NETWORK._serialized_start=871
  _NETWORK._serialized_end=970
  _VERSION._serialized_start=972
  _VERSION._serialized_end=1012
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Mapped convolution layers."""

from typing import Optional, Tuple, Union, List

from torch import Tensor, arange, cat, float64, int32, ones, split, no_grad
from torch.nn.functional import pad, unfold
from torch.nn.modules.conv import _ConvNd, Conv1d, Conv2d, Conv3d
from torch.nn.modules.utils import _single, _pair, _triple

from aihwkit.nn.functions import AnalogIndexedFunction, AnalogFunction
from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias
from aihwkit.simulator.tiles import BaseTile
from aihwkit.exceptions import ModuleError
from aihwkit.simulator.configs import SingleRPUConfig


class _AnalogConvNdMapped(AnalogModuleBase, _ConvNd):
    """Base class for convolution layers with tile mapping.

    """

    __constants__ = ['stride', 'padding', 'dilation', 'groups',
                     'padding_mode', 'output_padding', 'in_channels',
                     'out_channels', 'kernel_size', 'in_features', 'out_features',
                     'realistic_read_write',
                     'digital_bias', 'analog_bias', 'use_bias']
    in_channels: int
    out_channels: int
    kernel_size: Tuple[int, ...]
    stride: Tuple[int, ...]
    padding: Tuple[int, ...]
    dilation: Tuple[int, ...]
    realistic_read_write: bool
    transposed: bool
    output_padding: Tuple[int, ...]
    groups: int
    padding_mode: str
    fold_indices: Tensor
    input_size: float
    in_features: int
    out_features: int
    digital_bias: bool
    analog_bias: bool
    use_bias: bool
    use_indexed: Optional[bool]

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Tuple[int, ...],
            stride: Tuple[int, ...],
            padding: Tuple[int, ...],
            dilation: Tuple[int, ...],
            transposed: bool,
            output_padding: Tuple[int, ...],
            groups: int,
            bias: bool,
            padding_mode: str,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
            use_indexed: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments, too-many-locals
        if groups != 1:
            raise ValueError('Only one group is supported')
        if padding_mode != 'zeros':
            raise ValueError('Only "zeros" padding mode is supported')

        # Call super() after tile creation, including ``reset_parameters``.
        _ConvNd.__init__(self, in_channels, out_channels, kernel_size, stride,
                         padding, dilation, transposed, output_padding, groups, bias,
                         padding_mode)

        # Create tiles
        if rpu_config is None:
            rpu_config = SingleRPUConfig()

        rpu_config = self._set_weight_scaling_omega(rpu_config, weight_scaling_omega)

        AnalogModuleBase.__init__(
            self,
            self.get_tile_size(in_channels, groups, kernel_size),
            out_channels,
            bias,
            realistic_read_write,
            rpu_config.mapping
        )

        if self.analog_bias:
            raise ModuleError("AnalogConvNdMapped only supports digital bias.")

        if not rpu_config:
            rpu_config = SingleRPUConfig()

        max_input_size = rpu_config.mapping.max_input_size
        max_output_size = rpu_config.mapping.max_output_size
        kernel_elem = self.in_features // self.in_channels
        self.in_sizes = self.get_split_sizes(self.in_features, max_input_size, kernel_elem)
        self.out_sizes = self.get_split_sizes(self.out_features, max_output_size)

        self.analog_tile_array = []
        for i, in_tile_size in enumerate(self.in_sizes):
            in_tiles = []
            for j, out_tile_size in enumerate(self.out_sizes):
                tile = rpu_config.tile_class(out_tile_size,
                                             in_tile_size * kernel_elem,
                                             rpu_config,
                                             bias=self.analog_bias)
                self.register_analog_tile(tile, name=f"{i}_{j}")
                in_tiles.append(tile)
            self.analog_tile_array.append(in_tiles)

        # Set weights from the reset_parameters (since now the
        # analog_tiles are registered)
        self.set_weights(self.weight, self.bias)

        # Set the index matrices.
        self.use_indexed = use_indexed
        self.input_size = 0
        self.register_helper('input_size')
        self.fold_indices_lst = []  # type: List[Tensor]
        self.register_helper('fold_indices_lst')
        self.tensor_view = (-1,)  # type: Tuple[int, ...]

        # Unregister weight/bias as a parameter but keep it as a
        # field (needed for syncing still)
        self.unregister_parameter('weight')

    def get_split_sizes(self, size: int, split_max_size: int, group_size: int = 1) -> List[int]:
        """ Computed the split sizes across channels.

        Args:
            size: number of elements of the layer in one dimension
            split_max_size: max size of the split
            group_size: minimal size of features that needs to stay on one tile

        Returns:
            List of split sizes (in split groups)

        Raises:
            ModuleError: Tiling weight matrices is always done across channels
                only. If the group_size is larger than the
                maximal tile size, mapping cannot be done
        """
        if split_max_size <= 0:
            return [size // group_size]

        if group_size > split_max_size:
            raise ModuleError("Tile size too small to fit a single group (kernel): " +
                              f"{group_size} > {split_max_size}")

        size_per_group = size // group_size
        split_max_per_group = split_max_size // group_size

        n_splits = (size_per_group + split_max_per_group - 1) // split_max_per_group
        base, extra = divmod(size_per_group, n_splits)
        return [(base + (i < extra)) for i in range(n_splits)]

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        raise NotImplementedError

    def get_image_size(self, size: int, i: int) -> int:
        """Calculate the output image sizes."""
        nom = (size + 2 * self.padding[i] - self.dilation[i] * (self.kernel_size[i] - 1) - 1)
        return nom // self.stride[i] + 1

    def reset_parameters(self) -> None:
        """Reset the parameters (weight and bias)."""
        super().reset_parameters()
        if self.analog_tile_count():
            self.set_weights(self.weight, self.bias)

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        raise NotImplementedError

    def _recalculate_indexes(self, x_input: Tensor) -> None:
        """Calculate and set the indexes of the analog tile.

        Args:
            x_input: the input tensor.

        Raises:
            ModuleError: in case the input is not at least 3
            dimensional

        """
        self.input_size = x_input.numel() / x_input.size(0)
        if x_input.ndim < 3:
            raise ModuleError("Expect >2-dim inputs to convolutions")
        channel_dim = 1
        self.fold_indices_lst = []
        splits = split(x_input, self.in_sizes, dim=channel_dim)
        for x, in_channels, in_tiles in zip(splits, self.in_sizes, self.analog_tile_array):
            fold_indices, image_sizes, _ = self._calculate_indexes(x, in_channels)
            self.fold_indices_lst.append(fold_indices)

            for analog_tile in in_tiles:
                analog_tile.set_indexed(fold_indices, image_sizes)

    def _single_forward_indexed(self, analog_tile: BaseTile, x_input: Tensor) -> Tensor:
        """Compute the forward pass in indexed fashion. This is fast and
        memory-efficient indexed convolution (only for GPUs)"""

        return AnalogIndexedFunction.apply(
            analog_tile.get_analog_ctx(), x_input,
            analog_tile.shared_weights, not self.training)

    def _single_forward_unfold(self, analog_tile: BaseTile, x_input: Tensor) -> Tensor:
        """Forward using explicit unfolding (more suitable for CPUs) """
        im_shape = x_input.shape
        x_input_ = unfold(x_input, kernel_size=self.kernel_size, dilation=self.dilation,
                          padding=self.padding, stride=self.stride).transpose(1, 2)

        out = AnalogFunction.apply(
            analog_tile.get_analog_ctx(), x_input_,
            analog_tile.shared_weights, not self.training).transpose(1, 2)

        out_im_size = (im_shape[2] + 2 * self.padding[0]
                       - self.dilation[0] * (self.kernel_size[0] - 1) - 1) // self.stride[0] + 1
        return out.view(im_shape[0], analog_tile.out_size, out_im_size, -1)

    def forward(self, x_input: Tensor) -> Tensor:
        """Compute the forward pass."""
        # pylint: disable=arguments-differ, arguments-renamed, too-many-branches

        if self.use_indexed is None:
            use_indexed = self.analog_tile_array[0][0].device.type == 'cuda'
        else:
            use_indexed = self.use_indexed

        if use_indexed:
            input_size = x_input.numel() / x_input.size(0)
            if self.input_size != input_size or not self.analog_tile_array[0][0].is_indexed():
                self._recalculate_indexes(x_input)

        if self.analog_tile_count() == 1:
            analog_tile = self.analog_tile_array[0][0]

            if use_indexed:
                output = self._single_forward_indexed(analog_tile, x_input)
            else:
                output = self._single_forward_unfold(analog_tile, x_input)

            output = analog_tile.apply_out_scaling(output, self.tensor_view)

            if self.digital_bias:
                return output + self.bias.view(*self.tensor_view)
            return output

        # mapped version
        channel_dim = 1
        splits = split(x_input, self.in_sizes, dim=channel_dim)
        result = None  # type: Tensor
        for idx, (x, in_tiles) in enumerate(zip(splits, self.analog_tile_array)):
            out_result = []

            for analog_tile in in_tiles:
                if use_indexed:
                    output = self._single_forward_indexed(analog_tile, x)
                else:
                    output = self._single_forward_unfold(analog_tile, x)

                output = analog_tile.apply_out_scaling(output, self.tensor_view)
                out_result.append(output)

            if idx == 0:
                result = cat(out_result, channel_dim)
            else:
                result.add_(cat(out_result, channel_dim))

        # add bias to final result
        if self.digital_bias:
            return result + self.bias.view(*self.tensor_view)
        return result

    def set_weights(
            self,
            weight: Tensor,
            bias: Optional[Tensor] = None,
            force_exact: bool = False,
            apply_weight_scaling: bool = True,
            weight_scaling_omega: Optional[float] = None
    ) -> None:
        """Set the weight (and bias) with given Tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital alpha scale if
        ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.apply_weight_scaling`).

        Note:
            This is the recommended way for setting the weight/bias matrix of
            the analog tile, as it will correctly store the weights into the
            internal memory. Directly writing to ``self.weight`` and
            ``self.bias`` might yield wrong results as they are not always in
            sync with the analog tile Parameters, for performance reasons.

        Args:
            weight: weight matrix
            bias: bias vector
            force_exact: forces an exact write to the analog tiles
            apply_weight_scaling: Whether to rescale the given weight matrix
                and populate the digital output scaling factors as
                specified in the configuration
                :class:`~aihwkit.configs.utils.MappingParameter`. A
                new ``weight_scaling_omega`` can be given. Note that
                this will overwrite the existing digital out scaling
                factors.

                Note that each tile (in case of multiple mapped tiles)
                has it separate out scaling factors.

            weight_scaling_omega: The weight scaling omega factor (see
                :class:`~aihwkit.configs.utils.MappingParameter`). If
                given explicitly here, it will overwrite the value in
                the mapping field.
        """
        # pylint: disable=too-many-locals
        realistic = self.realistic_read_write and not force_exact

        shape = [self.out_features, self.in_channels, self.in_features // self.in_channels]
        weight = weight.clone().reshape(shape)

        weight_splits = split(weight, self.in_sizes, dim=1)

        for in_tiles, in_weight in zip(self.analog_tile_array, weight_splits):
            out_start = out_end = 0
            in_weight = in_weight.reshape([self.out_features, -1])
            for out_size, analog_tile in zip(self.out_sizes, in_tiles):
                out_end += out_size
                tile_weight = in_weight[out_start:out_end, :]
                out_start = out_end

                analog_tile.set_weights(
                    tile_weight, None,
                    apply_weight_scaling,
                    weight_scaling_omega
                )
                if realistic:
                    analog_tile.program_weights()

        self._sync_weights_from_tile()

        if self.digital_bias and bias is not None:
            with no_grad():
                self.bias.data[:] = bias[:]

    def get_weights(self, force_exact: bool = False,
                    apply_weight_scaling: bool = True) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the weight (and bias) tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital alpha scale if
        ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.apply_weight_scaling`).

        Note:
            This is the recommended way for setting the weight/bias matrix from
            the analog tile, as it will correctly fetch the weights from the
            internal memory. Accessing ``self.weight`` and ``self.bias`` might
            yield wrong results as they are not always in sync with the
            analog tile library, for performance reasons.

        Args:
            force_exact: forces an exact read to the analog tiles
            apply_weight_scaling: Whether to return the weights with the
                (digital) output scaling factors applied. Note the
                "logical" weights of the layer which the DNN is
                effectively using are those with the output scales
                applied. If ``apply_weight_scaling`` is set to False, then
                only the weight values that is programmed onto the
                crossbar array are returned, without applying the
                digital scales.

        Returns:
            tuple: weight matrix, bias vector
        """
        realistic = self.realistic_read_write and not force_exact

        weight_lst = []
        for in_tiles in self.analog_tile_array:
            in_tile_weight = []
            for analog_tile in in_tiles:
                if realistic:
                    tile_weight, _ = analog_tile.read_weights(apply_weight_scaling)
                else:
                    tile_weight, _ = analog_tile.get_weights(apply_weight_scaling)
                in_tile_weight.append(tile_weight)
            weight_lst.append(cat(in_tile_weight, 0))

        weight = cat(weight_lst, 1)

        if self.digital_bias:
            with no_grad():
                return weight, self.bias.data.clone().detach().cpu()
        return weight, None

    def extra_repr(self) -> str:
        """Set the extra representation of the module.

        Returns:
            A string with the extra representation.
        """
        output = AnalogModuleBase.extra_repr(self)
        output += ', mapping={}'.format((len(self.in_sizes), len(self.out_sizes)))

        return output


class AnalogConv1dMapped(_AnalogConvNdMapped):
    """1D convolution layer that maps to analog tiles.

    Applies a 1D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    The module will split the weight matrix onto multiple tiles if
    necessary. Physical max tile sizes are specified with
    :class:`~aihwkit.simulator.configs.utils.MappingParameter` in the
    RPU configuration, see
    :class:`~aihwkit.simulator.configs.configs.RPUConfigAlias`.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write for
            setting initial weights and read out of weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)

        if dilation != _single(1):
            raise ValueError('Only dilation = 1 is supported')

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _single(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, True
        )

        self.tensor_view = (-1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv1d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv1dMapped':
        """Return an AnalogConv1dMapped layer from a torch Conv1d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv1d layer based on the digital Conv1d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write,
                            )

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * kernel_size[0]

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCHW order?
        fold_indices = arange(2, x_input.size(2) + 2, dtype=float64).detach()
        shape = [1] + [1] + list(x_input.shape[2:])
        fold_indices = fold_indices.reshape(*shape)
        if not all(item == 0 for item in self.padding):
            fold_indices = pad(fold_indices, pad=[self.padding[0], self.padding[0]],
                               mode='constant', value=0)
        unfolded = fold_indices.unfold(2, self.kernel_size[0], self.stride[0]).clone()

        fold_indices = unfolded.reshape(-1, self.kernel_size[0]).transpose(0, 1).flatten().round()

        # concatenate the matrix index for different channels
        fold_indices_orig = fold_indices.clone()
        for i in range(in_channels - 1):
            fold_indices_tmp = fold_indices_orig.clone()
            for j in range(fold_indices_orig.size(0)):
                if fold_indices_orig[j] != 0:
                    fold_indices_tmp[j] += (input_size / in_channels) * (i + 1)

            fold_indices = cat([fold_indices, fold_indices_tmp], dim=0).clone()

        fold_indices = fold_indices.to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_height = x_input.size(2)
        d_height = self.get_image_size(x_height, 0)

        image_sizes = [in_channels, x_height, d_height]
        return (fold_indices, image_sizes, input_size)


class AnalogConv2dMapped(_AnalogConvNdMapped):
    """2D convolution layer that maps to analog tiles.

    Applies a 2D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    The module will split the weight matrix onto multiple tiles if
    necessary. Physical max tile sizes are specified with
    :class:`~aihwkit.simulator.configs.utils.MappingParameter` in the
    RPU configuration, see
    :class:`~aihwkit.simulator.configs.configs.RPUConfigAlias`.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
        use_indexed: Whether to use explicit unfolding or implicit indexing. If
            None (default), it will use implicit indexing for CUDA and
            explicit unfolding for CPU
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
            use_indexed: Optional[bool] = None,
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _pair(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, use_indexed
        )

        self.tensor_view = (-1, 1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv2d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv2dMapped':
        """Return an AnalogConv2dMapped layer from a torch Conv2d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv2dMapped layer based on the digital Conv2d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write)

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * kernel_size[0] * kernel_size[1]

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCHW order
        fold_indices = arange(2, input_size + 2, dtype=float64).detach()
        shape = [1] + list(x_input.shape[1:])
        fold_indices = fold_indices.reshape(*shape)
        fold_indices = unfold(fold_indices,
                              kernel_size=self.kernel_size,
                              stride=self.stride,
                              padding=self.padding,
                              dilation=self.dilation).flatten().round().to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0] * self.kernel_size[1])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_height = x_input.size(2)
        x_width = x_input.size(3)

        d_height = self.get_image_size(x_height, 0)
        d_width = self.get_image_size(x_width, 1)

        image_sizes = [in_channels, x_height, x_width, d_height, d_width]
        return (fold_indices, image_sizes, input_size)


class AnalogConv3dMapped(_AnalogConvNdMapped):
    """3D convolution layer that maps to analog tiles.

    Applies a 3D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    The module will split the weight matrix onto multiple tiles if
    necessary. Physical max tile sizes are specified with
    :class:`~aihwkit.simulator.configs.utils.MappingParameter` in the
    RPU configuration, see
    :class:`~aihwkit.simulator.configs.configs.RPUConfigAlias`.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling

    Raises:
        ModuleError: Tiling weight matrices is always done across channels
            only. If the kernel number of elements is larger than the
            maximal tile size, mapping cannot be done

    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        dilation = _triple(dilation)

        if dilation != _triple(1):
            raise ValueError('Only dilation = 1 is supported')

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _triple(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, True
        )

        self.tensor_view = (-1, 1, 1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv3d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv3dMapped':
        """Return an AnalogConv3dMapped layer from a torch Conv3d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv3d layer based on the digital Conv3d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write)

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * (
                kernel_size[0] * kernel_size[1] * kernel_size[2])

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        # pylint: disable=too-many-locals
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCDHW order
        fold_indices = arange(2, x_input.size(2) * x_input.size(3) * x_input.size(4) + 2,
                              dtype=float64).detach()
        shape = [1] + [1] + list(x_input.shape[2:])
        fold_indices = fold_indices.reshape(*shape)
        if not all(item == 0 for item in self.padding):
            fold_indices = pad(fold_indices, pad=[
                self.padding[2], self.padding[2],
                self.padding[1], self.padding[1],
                self.padding[0], self.padding[0]], mode='constant', value=0)
        unfolded = fold_indices.unfold(2, self.kernel_size[0], self.stride[0]). \
            unfold(3, self.kernel_size[1], self.stride[1]). \
            unfold(4, self.kernel_size[2], self.stride[2]).clone()

        fold_indices = unfolded.reshape(-1, self.kernel_size[0] * self.kernel_size[1] *
                                        self.kernel_size[2]).transpose(0, 1).flatten().round()

        # concatenate the matrix index for different channels
        fold_indices_orig = fold_indices.clone()
        for i in range(in_channels - 1):
            fold_indices_tmp = fold_indices_orig.clone()
            for j in range(fold_indices_orig.size(0)):
                if fold_indices_orig[j] != 0:
                    fold_indices_tmp[j] += (input_size / in_channels) * (i + 1)

            fold_indices = cat([fold_indices, fold_indices_tmp], dim=0).clone()

        fold_indices = fold_indices.to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0] *
                                                      self.kernel_size[1] *
                                                      self.kernel_size[2])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_depth = x_input.size(2)
        x_height = x_input.size(3)
        x_width = x_input.size(4)

        d_depth = self.get_image_size(x_depth, 0)
        d_height = self.get_image_size(x_height, 1)
        d_width = self.get_image_size(x_width, 2)

        image_sizes = [in_channels, x_depth, x_height, x_width, d_depth, d_height, d_width]
        return (fold_indices, image_sizes, input_size)
